import typing

import numpy as np
import pandas as pd
from alphabase.constants._const import PEAK_INTENSITY_DTYPE, PEAK_MZ_DTYPE
from alphabase.constants.atom import MASS_ISOTOPE
from alphabase.constants.modification import MOD_MASS
from alphabase.peptide.fragment import create_fragment_mz_dataframe, flatten_fragments
from alphabase.peptide.precursor import (
    calc_precursor_isotope_intensity,
    update_precursor_mz,
)

from alpharaw.match.match_utils import match_closest_peaks, match_highest_peaks


def make_query_df_for_peptide(
    sequence: str,
    mods: str,
    mod_sites: str,
    charge: int,
    rt_sec: float = 0.0,
    mobility: float = 0.0,
    ms_level: int = 2,
    charged_frag_types: list = ["b_z1", "b_z2", "y_z1", "y_z2"],
    include_fragments: bool = True,
    fragment_intensity_df: pd.DataFrame = None,
    include_precursor_isotopes: bool = False,
    max_isotope: int = 6,
    min_frag_mz: float = 100.0,
    min_frag_intensity: float = 0.001,
) -> pd.DataFrame:
    """
    Create a query dataframe for a peptide that can be
    use to generate MS2 or XIC plots.

    Parameters
    ----------
    sequence : str
        Peptide sequence.
    mods : str
        Modification names in alphabase format.
    mod_sites : str
        Modification sites in alphabase format.
    charge : int
        Charge state of the peptide.
    rt_sec : float, optional
        Retention time in seconds of the peptide, by default 0.0
    mobility : float, optional
        Ion mobility of the peptide, by default 0.0
    ms_level : int, optional
        MS level. If it is 1, `include_fragments` is always False. By default 2
    charged_frag_types : list, optional
        Charge states of fragments, by default ["b_z1", "b_z2", "y_z1", "y_z2"]
    include_fragments : bool, optional
        If include fragments in plot_df, by default True
    fragment_intensity_df : pd.DataFrame, optional
        Fragment intensity dataframe of the peptide for mirror plot, by default None
    include_precursor_isotopes : bool, optional
        If plot_df include precursor isotopes, by default False
    max_isotope : int, optional
        Maximal number of precursor isotopes, by default 6
    min_frag_mz : float, optional
        Minimal m/z value of fragments, by default 100.0
    min_frag_intensity : float, optional
        Minimal intensity of fragments to plot, by default 0.001

    Returns
    -------
    pd.DataFrame
        Query dataframe for plotting with possible columns:
        'mz', 'type', 'loss_type', 'charge', 'number', 'fragment_site',
       'ion_name', 'sequence', 'mods', 'mod_sites', 'precursor_charge',
       'modified_sequence', 'rt_sec', 'precursor_mz', 'precursor_i_0',
       'precursor_i_1', 'precursor_i_2', 'precursor_i_3', 'precursor_i_4',
       'precursor_i_5', 'precursor_mono_idx'
    """
    if ms_level == 1:
        include_fragments = False
    precursor_df, fragment_mz_df = make_precursor_fragment_df(
        sequence,
        mods,
        mod_sites,
        charge,
        include_fragments=include_fragments,
        charged_frag_types=charged_frag_types,
        include_precursor_isotopes=include_precursor_isotopes,
        max_isotope=max_isotope,
    )
    if fragment_intensity_df is not None:
        columns = np.intersect1d(
            fragment_mz_df.columns.values,
            fragment_intensity_df.columns.values,
        )
        fragment_mz_df = fragment_mz_df[columns]
        fragment_intensity_df = fragment_intensity_df[columns]

    return translate_frag_df_to_plot_df(
        precursor_df,
        fragment_mz_df,
        fragment_intensity_df=fragment_intensity_df,
        rt_sec=rt_sec,
        mobility=mobility,
        ms_level=ms_level,
        min_frag_mz=min_frag_mz,
        min_frag_intensity=min_frag_intensity,
    )


def make_psm_plot_df_for_peptide(
    spec_masses: np.ndarray,
    spec_intensities: np.ndarray,
    sequence: str,
    mods: str,
    mod_sites: str,
    charge: int,
    rt_sec: float = 0.0,
    mobility: float = 0.0,
    ppm: float = 20.0,
    charged_frag_types: list = ["b_z1", "b_z2", "y_z1", "y_z2"],
    include_fragments: bool = True,
    fragment_intensity_df: pd.DataFrame = None,
    include_precursor_isotopes: bool = False,
    max_isotope: int = 6,
    min_frag_mz: float = 100.0,
    min_frag_intensity: float = 0.001,
    match_mode: typing.Literal["closest", "highest"] = "closest",
) -> pd.DataFrame:
    """
    Create a plot dataframe for a MS2 spectrum and a peptide.

    Parameters
    ----------
    spec_masses : np.ndarray
        Peak m/z values of a spectrum.
    spec_intensities : np.ndarray
        Peak intensities of a spectrum.
    sequence : str
        Peptide sequence.
    mods : str
        Modification names in alphabase format.
    mod_sites : str
        Modification sites in alphabase format.
    charge : int
        Charge state of the peptide.
    rt_sec : float, optional
        Retention time in seconds of the peptide, by default 0.0
    mobility : float, optional
        Ion mobility of the peptide, by default 0.0
    ppm : float, optional
        Matching mass tolerance in ppm, by default 20.0
    charged_frag_types : list, optional
        Charge states of fragments, by default ["b_z1", "b_z2", "y_z1", "y_z2"]
    include_fragments : bool, optional
        If include fragments in plot_df, by default True
    fragment_intensity_df : pd.DataFrame, optional
        Fragment intensity dataframe of the peptide for mirror plot, by default None
    include_precursor_isotopes : bool, optional
        If plot_df include precursor isotopes, by default False
    max_isotope : int, optional
        Maximal number of precursor isotopes, by default 6
    min_frag_mz : float, optional
        Minimal m/z value of fragments, by default 100.0
    min_frag_intensity : float, optional
        Minimal intensity of fragments to plot, by default 0.001
    match_mode : "closest", "highest", optional
        If extract the closest peak or highest peak within
        the given matching tolerance, by default "closest"

    Returns
    -------
    pd.DataFrame
        Plot dataframe with possible columns:
        "modified_sequence", "mz", "intensity", "fragment_site",
        "ppm_err", "mass_err", "ion_name", "mod_sites",
        "precursor_mz", "precursor_i_0",
        "precursor_i_1", "precursor_i_2", "precursor_i_3", "precursor_i_4",
        "precursor_i_5", "precursor_mono_idx", "color"
    """
    plot_df = make_query_df_for_peptide(
        sequence,
        mods,
        mod_sites,
        charge,
        rt_sec=rt_sec,
        mobility=mobility,
        charged_frag_types=charged_frag_types,
        include_fragments=include_fragments,
        fragment_intensity_df=fragment_intensity_df,
        include_precursor_isotopes=include_precursor_isotopes,
        max_isotope=max_isotope,
        min_frag_mz=min_frag_mz,
        min_frag_intensity=min_frag_intensity,
    )

    return make_psm_plot_df(
        spec_masses=spec_masses,
        spec_intensities=spec_intensities,
        query_masses=plot_df.mz.values,
        query_ion_names=plot_df.ion_name.values,
        query_mass_tols=plot_df.mz.values * ppm * 1e-6,
        query_frag_idxes=plot_df.fragment_site.values,
        modified_sequence=plot_df.modified_sequence.values[0],
        mod_sites=mod_sites,
        query_intensities=plot_df.intensity.values
        if "intensity" in plot_df.columns
        else None,
        match_mode=match_mode,
    )


def make_psm_plot_df_for_frag_dfs(
    spec_masses: np.ndarray,
    spec_intensities: np.ndarray,
    precursor_df: pd.DataFrame,
    fragment_mz_df: pd.DataFrame,
    fragment_intensity_df: pd.DataFrame = None,
    ppm: float = 20.0,
    min_frag_mz: float = 100.0,
    min_frag_intensity: float = 0.001,
    match_mode: typing.Literal["closest", "highest"] = "closest",
) -> pd.DataFrame:
    """
    Similar to :func:`make_psm_plot_df_for_peptide`, but this function
    does not start from a peptide, but from its fragment_mz_df.

    Parameters
    ----------
    spec_masses : np.ndarray
        Peak m/z values of a spectrum.
    spec_intensities : np.ndarray
        Peak intensities of a spectrum.
    precursor_df : pd.DataFrame
        This must be alphabase precursor_df with only one precursor.
    fragment_mz_df : pd.DataFrame
        The alphabase fragment_mz_df of the `precursor_df`.
    fragment_intensity_df : pd.DataFrame, optional
        Fragment intensity dataframe of precursor for mirror plot, by default None
    ppm : float, optional
        Matching mass tolerance in ppm, by default 20.0
    min_frag_mz : float, optional
        Minimal m/z value of fragments, by default 100.0
    min_frag_intensity : float, optional
        Minimal intensity of fragments to plot, by default 0.001
    match_mode : "closest", "highest", optional
        If extract the closest peak or highest peak within
        the given matching tolerance, by default "closest"

    Returns
    -------
    DataFrame
        Plot dataframe with possible columns:
        "modified_sequence", "mz", "intensity", "fragment_site",
        "ppm_err", "mass_err", "ion_name", "mod_sites",
        "precursor_mz", "color"
    """
    plot_df = translate_frag_df_to_plot_df(
        precursor_df,
        fragment_mz_df,
        fragment_intensity_df=fragment_intensity_df,
        min_frag_mz=min_frag_mz,
        min_frag_intensity=min_frag_intensity,
    )

    if "intensity" in plot_df.columns:
        query_intensities = plot_df.intensity.values
    else:
        query_intensities = None

    return make_psm_plot_df(
        spec_masses=spec_masses,
        spec_intensities=spec_intensities,
        query_masses=plot_df.mz.values,
        query_ion_names=plot_df.ion_name.values,
        query_mass_tols=plot_df.mz.values * ppm * 1e-6,
        query_frag_idxes=plot_df.fragment_site.values,
        modified_sequence=plot_df.modified_sequence.values[0],
        mod_sites=precursor_df.mod_sites.values[0],
        query_intensities=query_intensities,
        match_mode=match_mode,
    )


def make_query_df(
    query_masses: np.ndarray,
    query_ion_names: typing.List[str],
    query_rt_sec: float,
    precursor_mz: float,
    query_im: float = 0.0,
    query_intensities: np.ndarray = None,
) -> pd.DataFrame:
    """
    Create a query dataframe based on query_masses and query_ion_names
    to generate MS2 or XIC plots.

    Parameters
    ----------
    query_masses : np.ndarray
        Query or fragment m/z values.
    query_ion_names : typing.List[str]
        Query or fragment ion names.
    query_rt_sec : float
        RT in seconds.
    precursor_mz : float
        Precursor m/z value.
    query_im : float, optional
        Ion mobility, by default 0.0
    query_intensities : np.ndarray, optional
        Query intensities for mirror plot, by default None

    Returns
    -------
    pd.DataFrame
        "mz", "intensity", "fragment_site", "ion_name", "precursor_mz"
    """
    df = pd.DataFrame(
        dict(
            mz=query_masses,
            ion_name=query_ion_names,
            rt_sec=query_rt_sec,
            precursor_mz=precursor_mz,
        )
    )
    if query_im:
        df["im"] = query_im
    if query_intensities is not None:
        df["intensity"] = query_intensities
    return df


def make_psm_plot_df(
    spec_masses: np.ndarray,
    spec_intensities: np.ndarray,
    query_masses: np.ndarray,
    query_ion_names: typing.List[str],
    query_mass_tols: np.ndarray,
    query_frag_idxes: np.ndarray,
    modified_sequence: str = "",
    mod_sites: str = "",
    query_intensities: np.ndarray = None,
    match_mode: typing.Literal["closest", "highest"] = "closest",
) -> pd.DataFrame:
    """
    Create a plot dataframe for a MS2 spectrum and query m/z values.

    Parameters
    ----------
    spec_masses : np.ndarray
        Peak m/z values of a spectrum.
    spec_intensities : np.ndarray
        Peak intensities of a spectrum.
    query_masses : np.ndarray
        Query or fragment m/z values.
    query_ion_names : typing.List[str]
        Query or fragment ion names.
    query_mass_tols : np.ndarray
        Matching mass tolerance of `query_masses`.
    query_frag_idxes : np.ndarray
        Fragment indices or positions of `query_masses`.
    modified_sequence : str, optional
        Modified sequence, such as "AM[+16]DEFGK_(2+)", by default ""
    mod_sites : str, optional
        Modification sites in alphabase format, by default ""
    query_intensities : np.ndarray, optional
        Query intensities for mirror plot, by default None
    match_mode : "closest", "highest", optional
        If extract the closest peak or highest peak within
        the given matching tolerance, by default "closest"

    Returns
    -------
    pd.DataFrame
        Plot dataframe with possible columns:
        "modified_sequence", "mz", "intensity", "fragment_site",
        "ppm_err", "mass_err", "ion_name", "mod_sites",
        "precursor_mz", "precursor_i_0",
        "precursor_i_1", "precursor_i_2", "precursor_i_3", "precursor_i_4",
        "precursor_i_5", "precursor_mono_idx", "color"
    """
    query_ion_names = np.array(query_ion_names, dtype="U")
    if match_mode == "highest":
        query_matched_idxes = match_highest_peaks(
            spec_masses,
            spec_intensities,
            query_mzs=query_masses,
            query_mz_tols=query_mass_tols,
        )
    else:
        query_matched_idxes = match_closest_peaks(
            spec_masses,
            spec_intensities,
            query_mzs=query_masses,
            query_mz_tols=query_mass_tols,
        )
    matched_bools = (query_matched_idxes != -1) & (query_masses > 0)
    matched_query_masses = query_masses[matched_bools]
    matched_frag_idxes = query_frag_idxes[matched_bools]
    matched_ion_names = query_ion_names[matched_bools]
    matched_idxes = query_matched_idxes[matched_bools]
    matched_spec_masses = spec_masses[matched_idxes]
    matched_spec_intens = spec_intensities[matched_idxes]
    matched_mass_errs = matched_spec_masses - matched_query_masses
    ppm_mass_errs = matched_mass_errs * 1e6 / matched_query_masses
    query_spec_intens = spec_intensities[query_matched_idxes]
    query_spec_intens[~matched_bools] = 0

    spec_df = pd.DataFrame(
        dict(
            modified_sequence=modified_sequence,
            mz=spec_masses,
            intensity=spec_intensities,
            fragment_site=-1,
            ppm_err=0,
            mass_err=0,
            ion_name="-",
        )
    )
    matched_df = pd.DataFrame(
        dict(
            modified_sequence=modified_sequence,
            mz=matched_query_masses,
            intensity=matched_spec_intens,
            fragment_site=matched_frag_idxes,
            ppm_err=ppm_mass_errs,
            mass_err=matched_mass_errs,
            ion_name=matched_ion_names,
        )
    )

    if query_intensities is None or len(matched_df) == 0:
        mirror_df = pd.DataFrame()
    else:
        query_intensities *= -(
            matched_spec_intens.max() / query_intensities[matched_bools].max()
        )
        mirror_df = pd.DataFrame(
            dict(
                modified_sequence=modified_sequence,
                mz=query_masses,
                intensity=query_intensities,
                fragment_site=-1,
                ppm_err=0,
                mass_err=0,
                ion_name=query_ion_names,
            )
        )

        def PCC_sim(x, y):
            x = x - np.mean(x)
            y = y - np.mean(y)
            return np.dot(x, y) / (np.linalg.norm(x) * np.linalg.norm(y) + 1e-8)

        PCC = PCC_sim(-query_intensities, query_spec_intens)

    df = pd.concat([spec_df, matched_df, mirror_df], ignore_index=True)
    df["mod_sites"] = mod_sites
    if len(mirror_df) > 0:
        df["pcc"] = PCC
    return df


def get_modified_sequence(
    sequence: str, mods: str, mod_sites: str, charge: int = 0
) -> str:
    """
    Parse sequence, mods, mod_sites into a single modified sequence string.

    Parameters
    ----------
    sequence : str
        Peptide sequence.
    mods : str
        Modifications in alphabase format.
    mod_sites : str
        Modification sites in alphabase format.
    charge : int, optional
        Precursor charge, by default 0

    Returns
    -------
    str
        Modified sequence.
    """
    sequence = "_" + sequence + "_"
    mod_masses = np.zeros(len(sequence))
    if mods:
        for mod, mod_site in zip(mods.split(";"), mod_sites.split(";")):
            mod_masses[int(mod_site)] += MOD_MASS[mod]
        mod_masses = np.rint(mod_masses)
    mod_seq = sequence
    for i in range(len(sequence) - 1, -1, -1):
        if mod_masses[i] != 0:
            mod_seq = mod_seq[: i + 1] + f"[{mod_masses[i]:+.0f}]" + mod_seq[i + 1 :]
    if charge != 0:
        return mod_seq + f"({charge}+)"
    else:
        return mod_seq


def translate_frag_df_to_plot_df(
    precursor_df: pd.DataFrame,
    fragment_mz_df: pd.DataFrame,
    fragment_intensity_df: pd.DataFrame = None,
    rt_sec: float = 0.0,
    mobility: float = 0.0,
    ms_level: int = 2,
    min_frag_mz: float = 100.0,
    min_frag_intensity: float = 0.001,
) -> pd.DataFrame:
    """
    Translate `precursor_df`, `fragment_mz_df` (and `fragment_intensity_df`)
    into a single plot df.

    Parameters
    ----------
    precursor_df : pd.DataFrame
        Precursor df with a single precursor.
    fragment_mz_df : pd.DataFrame
        fragment_mz_df for the precursor.
    fragment_intensity_df : pd.DataFrame, optional
        fragment_intensity_df for the precursor, by default None
    rt_sec : float, optional
        RT in seconds, by default 0.0
    mobility : float, optional
        Ion mobility, by default 0.0
    ms_level : int, optional
        MS level. If 2, precursor_mz will be included otherwise it is useless.
        By default 2
    min_frag_mz : float, optional
        Minimal m/z value of fragments, by default 100.0
    min_frag_intensity : float, optional
        Minimal intensity of fragments to plot, by default 0.001

    Returns
    -------
    pd.DataFrame
        Query dataframe for plotting with possible columns:
        'mz', 'type', 'loss_type', 'charge', 'number', 'fragment_site',
       'ion_name', 'sequence', 'mods', 'mod_sites', 'precursor_charge',
       'modified_sequence', 'rt_sec', 'precursor_mz'
    """
    fragment_mz_df = fragment_mz_df.mask(fragment_mz_df < min_frag_mz, 0)
    if fragment_intensity_df is None:
        fragment_intensity_df = pd.DataFrame()

    flat_columns = ["type", "number", "loss_type", "charge"]
    if len(fragment_mz_df) > 0:
        precursor_df, fragment_df = flatten_fragments(
            precursor_df,
            fragment_mz_df,
            fragment_intensity_df,
            min_fragment_intensity=min_frag_intensity,
            custom_columns=flat_columns + ["position"],
        )

        fragment_df.rename(columns={"position": "fragment_site"}, inplace=True)

        fragment_df["ion_name"] = fragment_df[flat_columns].apply(
            lambda x: chr(x.iloc[0])
            + str(x.iloc[1])
            + (f"{-x.iloc[2]:+}" if x.iloc[2] != 0 else "")
            + "+" * x.iloc[3],
            axis=1,
        )
    else:
        fragment_df = pd.DataFrame(
            columns=["mz", "intensity", "type", "fragment_site", "ion_name"]
            + flat_columns
        )

    isotope_names = [col for col in precursor_df.columns if col.startswith("i_")]
    if len(isotope_names) > 0:
        isotope_df = pd.DataFrame(
            np.zeros(
                (len(isotope_names), len(fragment_df.columns)),
            ),
            columns=fragment_df.columns,
        )
        isotope_df["type"] = ord("M")
        charge = precursor_df.charge.values[0]
        precursor_mz = precursor_df.precursor_mz.values[0]
        isotope_mzs = []
        ion_names = []
        mono_idx = precursor_df.mono_isotope_idx.values[0]
        for i in range(len(isotope_names)):
            isotope_mzs.append(precursor_mz + (i - mono_idx) * MASS_ISOTOPE / charge)
            ion_names.append(f"M{i-mono_idx}")
        isotope_df["mz"] = np.array(isotope_mzs, dtype=PEAK_MZ_DTYPE)
        if "intensity" in fragment_df.columns:
            isotope_df["intensity"] = precursor_df[isotope_names].values.astype(
                PEAK_INTENSITY_DTYPE
            )[0, :]
        isotope_df["ion_name"] = ion_names
        isotope_df["fragment_site"] = -1
        fragment_df = pd.concat((fragment_df, isotope_df), ignore_index=True)

    fragment_df["sequence"] = precursor_df.sequence.values[0]
    fragment_df["mods"] = precursor_df.mods.values[0]
    fragment_df["mod_sites"] = precursor_df.mod_sites.values[0]
    fragment_df["precursor_charge"] = precursor_df.charge.values[0]
    fragment_df["modified_sequence"] = get_modified_sequence(
        sequence=precursor_df.sequence.values[0],
        mods=precursor_df.mods.values[0],
        mod_sites=precursor_df.mod_sites.values[0],
        charge=0
        if "charge" not in precursor_df.columns
        else precursor_df.charge.values[0],
    )
    if rt_sec:
        fragment_df["rt_sec"] = rt_sec
    if mobility:
        fragment_df["im"] = mobility

    if ms_level == 2:
        fragment_df["precursor_mz"] = precursor_df.precursor_mz.values[0]
        for iso in isotope_names:
            fragment_df[f"precursor_{iso}"] = precursor_df[iso].values[0]
        if len(isotope_names) > 0:
            fragment_df["precursor_mono_idx"] = precursor_df.mono_isotope_idx.values[0]

    return fragment_df


def make_precursor_fragment_df(
    sequence: str,
    mods: str,
    mod_sites: str,
    charge: int,
    include_fragments: bool = True,
    charged_frag_types: list = ["b_z1", "b_z2", "y_z1", "y_z2"],
    include_precursor_isotopes: bool = False,
    max_isotope: int = 6,
) -> typing.Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Create precursor_df and fragment_mz_df for a peptide.

    Parameters
    ----------
    sequence : str
        Peptide sequence
    mods : str
        Modifications in alphabase format
    mod_sites : str
        Modification sites in alphabase format
    charge : int
        Precursor charge state
    include_fragments : bool, optional
        If calculate fragments (fragment_mz_df), by default True
    charged_frag_types : list, optional
        Fragment charge states, by default ["b_z1", "b_z2", "y_z1", "y_z2"]
    include_precursor_isotopes : bool, optional
        If calculate precursor isotopes, by default False
    max_isotope : int, optional
        Maximal number of isotopes, by default 6

    Returns
    -------
    Tuple
        pd.DataFrame: precursor dataframe.
        pd.DataFrame: fragment dataframe. Empty dataframe if include_fragments==False
    """
    precursor_df = pd.DataFrame(
        dict(sequence=[sequence], mods=[mods], mod_sites=[mod_sites], charge=charge)
    )
    update_precursor_mz(precursor_df)
    if include_precursor_isotopes:
        calc_precursor_isotope_intensity(precursor_df, max_isotope=max_isotope)
    if include_fragments:
        fragment_mz_df = create_fragment_mz_dataframe(precursor_df, charged_frag_types)
    else:
        precursor_df["frag_start_idx"] = 0
        precursor_df["frag_stop_idx"] = 0
        fragment_mz_df = pd.DataFrame(columns=charged_frag_types)
    return precursor_df, fragment_mz_df



import numpy as np
import pandas as pd
from alphatims.utils import threadpool
from numba import njit

from alpharaw.feature.centroids import connect_centroids


@threadpool(include_progress_callback=False)
@njit
def path_finder(
    x: np.ndarray,
    from_idx: np.ndarray,
    to_idx: np.ndarray,
    forward: np.ndarray,
    backward: np.ndarray,
):
    """Extracts path information and writes to path matrix.

    Args:
        x (np.ndarray): Input index. Note that we are using the performance function so this is a range.
        from_idx (np.ndarray): Array containing from indices.
        to_idx (np.ndarray): Array containing to indices.
        forward (np.ndarray): Array to report forward connection.
        backward (np.ndarray): Array to report backward connection.
    """

    fr = from_idx[x]
    to = to_idx[x]

    forward[fr] = to
    backward[to] = fr


@threadpool(include_progress_callback=False)
@njit
def find_path_start(
    x: np.ndarray, forward: np.ndarray, backward: np.ndarray, path_starts: np.ndarray
):
    """Function to find the start of a path.

    Args:
        x (np.ndarray): Input index. Note that we are using the performance function so this is a range.
        forward (np.ndarray):  Array to report forward connection.
        backward (np.ndarray):  Array to report backward connection.
        path_starts (np.ndarray): Array to report path starts.
    """
    if forward[x] > -1 and backward[x] == -1:
        path_starts[x] = 0


@threadpool(include_progress_callback=False)
@njit
def find_path_length(
    x: np.ndarray, path_starts: np.ndarray, forward: np.ndarray, path_cnt: np.ndarray
):
    """Function to extract the length of a path.

    Args:
        x (np.ndarray): Input index. Note that we are using the performance function so this is a range.
        path_starts (np.ndarray): Array that stores the starts of the paths.
        forward (np.ndarray): Array that stores forward information.
        path_cnt (np.ndarray): Reporting array to count the paths.
    """
    ctr = 1
    idx = path_starts[x]
    while forward[idx] > -1:
        ctr += 1
        idx = forward[idx]
    path_cnt[x] = ctr


@threadpool(include_progress_callback=False)
@njit
def fill_path_matrix(
    x: np.ndarray,
    path_start: np.ndarray,
    forwards: np.ndarray,
    out_hill_data: np.ndarray,
    out_hill_ptr: np.ndarray,
):
    """Function to fill the path matrix.

    Args:
        x (np.ndarray): Input index. Note that we are using the performance function so this is a range.
        path_starts (np.ndarray): Array that stores the starts of the paths.
        forwards (np.ndarray): Forward array.
        out_hill_data (np.ndarray): Array containing the indices to hills.
        out_hill_ptr (np.ndarray): Array containing the bounds to out_hill_data.
    """
    path_position = 0
    idx = path_start[x]
    while idx > -1:
        out_hill_data[out_hill_ptr[x] + path_position] = idx
        idx = forwards[idx]
        path_position += 1


def get_hills(
    centroids: np.ndarray,
    from_idx: np.ndarray,
    to_idx: np.ndarray,
    hill_length_min: int = 3,
) -> (np.ndarray, np.ndarray, int):
    """Function to get hills from centroid connections.

    Args:
        centroids (np.ndarray): 1D Array containing the masses of the centroids.
        from_idx (np.ndarray): From index.
        to_idx (np.ndarray): To index.
        hill_length_min (int): Minimum hill length:

    Returns:
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        path_node_cnt (int): Number of elements in this path.
    """

    forward = np.full(centroids.shape[0], -1)
    backward = np.full(centroids.shape[0], -1)
    path_starts = np.full(centroids.shape[0], -1)

    path_finder(range(len(from_idx)), from_idx, to_idx, forward, backward)
    find_path_start(range(len(forward)), forward, backward, path_starts)

    # path_starts will now container the first index of all connected centroids
    path_starts = np.where(path_starts == 0)[0]

    path_node_cnt = np.full(path_starts.shape[0], -1)
    find_path_length(range(len(path_starts)), path_starts, forward, path_node_cnt)

    relavant_path_node = np.where(path_node_cnt >= hill_length_min)[0]
    path_starts = np.take(path_starts, relavant_path_node)
    path_node_cnt = np.take(path_node_cnt, relavant_path_node)
    del relavant_path_node

    # Generate the hill matix indice ptr data
    hill_ptrs = np.empty((path_starts.shape[0] + 1), dtype=np.int32)

    hill_ptrs[0] = 0
    hill_ptrs[1:] = path_node_cnt.cumsum()
    hill_data = np.empty((int(hill_ptrs[-1])), np.int32)

    fill_path_matrix(
        range(len(path_starts)), path_starts, forward, hill_data, hill_ptrs
    )

    del from_idx, to_idx, path_starts, forward, backward
    return hill_ptrs, hill_data, path_node_cnt


def extract_hills(
    query_data: dict, max_gap: int, centroid_tol: float
) -> (np.ndarray, np.ndarray, int, float, float):
    """[summary]

    Args:
        query_data (dict): Data structure containing the query data.
        max_gap (int): Maximum gap when connecting centroids.
        centroid_tol (float): Centroid tolerance.

    Returns:
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        path_node_cnt (int): Number of elements in this path.
        score_median (float): Median score.
        score_std (float): Std deviation of the score.
    """

    indices = np.array(query_data["indices_ms1"])
    mass_data = np.array(query_data["mass_list_ms1"])

    rowwise_peaks = indices[1:] - indices[:-1]
    row_borders = indices[1:]

    from_idx, to_idx, score_median, score_std = connect_centroids(
        rowwise_peaks, row_borders, mass_data, max_gap, centroid_tol
    )

    hill_ptrs, hill_data, path_node_cnt = get_hills(mass_data, from_idx, to_idx)

    del mass_data
    del indices

    return hill_ptrs, hill_data, path_node_cnt, score_median, score_std


@njit
def remove_duplicate_hills(hill_ptrs, hill_data, path_node_cnt):
    """
    Removes hills that share datapoints. Starts from the largest hills.

    """
    taken_points = np.zeros(hill_data.max() + 1)

    c = 0
    current_idx = 0

    hill_ptrs_new = np.zeros_like(hill_ptrs)
    hill_data_new = np.zeros_like(hill_data)

    for p in np.argsort(path_node_cnt)[::-1]:
        s, e = hill_ptrs[p], hill_ptrs[p + 1]

        point_idx = hill_data[s:e]

        hill_pts = taken_points[point_idx]

        if hill_pts.sum() == 0:
            hill_data_new[current_idx : current_idx + len(hill_pts)] = point_idx
            current_idx += len(hill_pts)
            hill_ptrs_new[c + 1] = current_idx
            c += 1

        taken_points[point_idx] += 1

    hill_data_new = hill_data_new[:current_idx]
    hill_ptrs_new = hill_ptrs_new[:c]

    return hill_ptrs_new, hill_data_new


@njit
def fast_minima(y: np.ndarray) -> np.ndarray:
    """Function to calculate the local minimas of an array.

    Args:
        y (np.ndarray): Input array.

    Returns:
        np.ndarray: Array containing minima positions.
    """
    minima = np.zeros(len(y))

    start = 0
    end = len(y)

    for i in range(start + 2, end - 2):
        if (
            ((y[i - 1] > y[i]) & (y[i + 1] > y[i]))
            or ((y[i - 1] > y[i]) & (y[i + 1] == y[i]) & (y[i + 2] > y[i]))
            or ((y[i - 2] > y[i]) & (y[i - 1] == y[i]) & (y[i + 1] > y[i]))
            or (
                (y[i - 2] > y[i])
                & (y[i - 1] == y[i])
                & (y[i + 1] == y[i])
                & (y[i + 2] > y[i])
            )
        ):
            minima[i] = 1

    minima = minima.nonzero()[0]

    return minima


# %% ../nbs/04_feature_finding.ipynb 15
@threadpool(include_progress_callback=False)
@njit
def split(
    k: np.ndarray,
    hill_ptrs: np.ndarray,
    int_data: np.ndarray,
    hill_data: np.ndarray,
    splits: np.ndarray,
    hill_split_level: float,
    window: int,
):
    """Function to split hills.

    Args:
        k (np.ndarray): Input index. Note that we are using the performance function so this is a range.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        hill_data (np.ndarray): Array containing the indices to hills.
        splits (np.ndarray): Array containing splits.
        hill_split_level (float): Split level for hills.
        window (int): Smoothing window.
    """

    start = hill_ptrs[k]
    end = hill_ptrs[k + 1]

    int_idx = hill_data[start:end]  # index to hill data

    int_trace = int_data[int_idx]

    for i in range(len(int_idx)):
        min_index = max(0, i - window)
        max_index = min(len(int_idx), i + window + 1)
        int_trace[i] = np.median(int_trace[min_index:max_index])

    for i in range(len(int_idx)):
        min_index = max(0, i - window)
        max_index = min(len(int_idx), i + window + 1)
        int_trace[i] = np.mean(int_trace[min_index:max_index])

    # minima = (np.diff(np.sign(np.diff(int_trace))) > 0).nonzero()[0] + 1 #This works also but is slower

    minima = fast_minima(int_trace)

    sorted_minima = np.argsort(int_trace[minima])

    minima = minima[sorted_minima]

    for min_ in minima:
        minval = int_trace[min_]

        left_max = max(int_trace[:min_])
        right_max = max(int_trace[min_:])

        min_max = min(left_max, right_max)

        if (minval == 0) or ((min_max / minval) > hill_split_level):
            splits[k] = start + min_
            break  # Split only once per iteration


def split_hills(
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
    int_data: np.ndarray,
    hill_split_level: float,
    window: int,
) -> np.ndarray:
    """Wrapper function to split hills

    Args:
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        hill_split_level (float): Split level for hills.
        window (int): Smoothing window.

    Returns:
        np.ndarray: Array containing the bounds to the hill_data with splits.
    """

    splits = np.zeros(len(int_data), dtype=np.int32)
    to_check = np.arange(len(hill_ptrs) - 1)

    while len(to_check) > 0:
        split(
            to_check, hill_ptrs, int_data, hill_data, splits, hill_split_level, window
        )
        splitpoints = splits.nonzero()[0]

        to_check = np.zeros(len(hill_ptrs))
        to_check[splitpoints] = 1

        to_check = np.insert(
            to_check, splitpoints + 1, np.ones(len(splitpoints))
        ).nonzero()[0]  # array, index, what
        hill_ptrs = np.insert(
            hill_ptrs, splitpoints + 1, splits[splitpoints]
        )  # array, index, what

        splits = np.zeros(len(hill_ptrs), dtype=np.int32)  # was np np.int32

    return hill_ptrs


# %% ../nbs/04_feature_finding.ipynb 17
@threadpool(include_progress_callback=False)
@njit
def check_large_hills(
    idx: np.ndarray,
    large_peaks: np.ndarray,
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
    int_data: np.ndarray,
    to_remove: np.ndarray,
    large_peak: int = 40,
    hill_peak_factor: float = 2,
    window: int = 1,
):
    """Function to check large hills and flag them for removal.

    Args:
        idx (np.ndarray): Input index. Note that we are using the performance function so this is a range.
        large_peaks (np.ndarray): Array containing large peaks.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        to_remove (np.ndarray): Array with indexes to remove.
        large_peak (int, optional): Length criterion when a peak is large. Defaults to 40.
        hill_peak_factor (float, optional): Hill maximum criterion. Defaults to 2.
        window (int, optional): Smoothing window.. Defaults to 1.
    """
    k = large_peaks[idx]

    start = hill_ptrs[k]
    end = hill_ptrs[k + 1]

    int_idx = hill_data[start:end]  # index to hill data

    int_smooth_ = int_data[int_idx]

    for i in range(len(int_idx)):
        min_index = max(0, i - window)
        max_index = min(len(int_idx), i + window + 1)
        int_smooth_[i] = np.median(int_smooth_[min_index:max_index])

    for i in range(len(int_idx)):
        min_index = max(0, i - window)
        max_index = min(len(int_idx), i + window + 1)
        int_smooth_[i] = np.mean(int_smooth_[min_index:max_index])

    int_ = int_data[int_idx]

    max_ = np.max(int_)

    if (max_ / int_smooth_[0] > hill_peak_factor) & (
        max_ / int_smooth_[-1] > hill_peak_factor
    ):
        to_remove[idx] = 0


def filter_hills(
    hill_data: np.ndarray,
    hill_ptrs: np.ndarray,
    int_data: np.ndarray,
    hill_check_large: int = 40,
    window: int = 1,
) -> (np.ndarray, np.ndarray):
    """Filters large hills.

    Args:
        hill_data (np.ndarray): Array containing the indices to hills.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        hill_check_large (int, optional): Length criterion when a hill is considered large.. Defaults to 40.
        window (int, optional): Smoothing window. Defaults to 1.

    Returns:
        np.ndarray: Filtered hill data.
        np.ndarray: Filtered hill points.
    """

    large_peaks = np.where(np.diff(hill_ptrs) >= hill_check_large)[0]

    to_remove = np.ones(len(large_peaks), dtype=np.int32)
    check_large_hills(
        range(len(large_peaks)),
        large_peaks,
        hill_ptrs,
        hill_data,
        int_data,
        to_remove,
        window,
    )

    idx_ = np.ones(len(hill_data), dtype=np.int32)
    keep = np.ones(len(hill_ptrs) - 1, dtype=np.int32)

    to_remove = to_remove.nonzero()[0]

    for _ in to_remove:
        idx_[hill_ptrs[_] : hill_ptrs[_ + 1]] = 0
        keep[_] = 0

    hill_lens = np.diff(hill_ptrs)
    keep_ = hill_lens[keep.nonzero()[0]]

    hill_data_ = hill_data[idx_.nonzero()[0]]
    hill_ptrs_ = np.empty((len(keep_) + 1), dtype=np.int32)
    hill_ptrs_[0] = 0
    hill_ptrs_[1:] = keep_.cumsum()

    return hill_data_, hill_ptrs_


# %% ../nbs/04_feature_finding.ipynb 20
@threadpool
@njit
def hill_stats(
    idx: np.ndarray,
    hill_range: np.ndarray,
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
    int_data: np.ndarray,
    mass_data: np.ndarray,
    rt_: np.ndarray,
    rt_idx: np.ndarray,
    stats: np.ndarray,
    hill_nboot_max: int,
    hill_nboot: int,
):
    """Function to calculate hill stats.

    Args:
        idx (np.ndarray): Input index. Note that we are using the performance function so this is a range.
        hill_range (np.ndarray): Hill range.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        mass_data (np.ndarray): Array containing mass data.
        rt_ (np.ndarray): Array with retention time information for each scan.
        rt_idx (np.ndarray): Lookup array to match centroid idx to rt.
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        hill_nboot_max (int): Maximum number of bootstrap comparisons.
        hill_nboot (int): Number of bootstrap comparisons
    """
    np.random.seed(42)

    start = hill_ptrs[idx]
    end = hill_ptrs[idx + 1]

    idx_ = hill_data[start:end]

    int_ = int_data[idx_]
    mz_ = mass_data[idx_]

    ms1_int_apex = np.max(int_)
    ms1_int_area = np.abs(np.trapz(int_, rt_[rt_idx[idx_]]))  # Area

    rt_min = rt_[rt_idx[idx_]].min()
    rt_max = rt_[rt_idx[idx_]].max()

    bootsize = hill_nboot_max if len(idx_) > hill_nboot_max else len(idx_)

    averages = np.zeros(hill_nboot)
    average = 0

    for i in range(hill_nboot):
        boot = np.random.choice(len(int_), bootsize, replace=True)
        boot_mz = np.sum(mz_[boot] * int_[boot]) / np.sum(int_[boot])
        averages[i] = boot_mz
        average += boot_mz

    average_mz = average / hill_nboot

    delta = 0
    for i in range(hill_nboot):
        delta += (average_mz - averages[i]) ** 2  # maybe easier?
    delta_m = np.sqrt(delta / (hill_nboot - 1))

    stats[idx, 0] = average_mz
    stats[idx, 1] = delta_m
    stats[idx, 2] = ms1_int_area
    stats[idx, 3] = ms1_int_apex
    stats[idx, 4] = rt_min
    stats[idx, 5] = rt_max


def remove_duplicates(
    stats: np.ndarray, hill_data: np.ndarray, hill_ptrs: np.ndarray
) -> (np.ndarray, np.ndarray, np.ndarray):
    """Remove duplicate hills.

    Args:
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        hill_data (np.ndarray): Array containing the indices to hills.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.

    Returns:
        np.ndarray: Filtered hill data.
        np.ndarray: Filtered hill points.
        np.ndarray: Filtered hill stats.
    """

    dups = pd.DataFrame(stats).duplicated()  # all duplicated hills

    idx_ = np.ones(len(hill_data), dtype=np.int32)  # keep all
    keep = np.ones(len(hill_ptrs) - 1, dtype=np.int32)

    for _ in np.arange(len(stats))[dups]:  # duplicates will be assigned zeros
        idx_[hill_ptrs[_] : hill_ptrs[_ + 1]] = 0
        keep[_] = 0

    hill_lens = np.diff(hill_ptrs)
    keep_ = hill_lens[keep.nonzero()[0]]

    hill_data_ = hill_data[idx_.nonzero()[0]]
    hill_ptrs_ = np.empty((len(keep_) + 1), dtype=np.int32)
    hill_ptrs_[0] = 0
    hill_ptrs_[1:] = keep_.cumsum()

    return hill_data_, hill_ptrs_, stats[~dups]


def get_hill_data(
    query_data: dict,
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
    hill_nboot_max: int = 300,
    hill_nboot: int = 150,
) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray):
    """Wrapper function to get the hill data.

    Args:
        query_data (dict): Data structure containing the query data.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        hill_nboot_max (int): Maximum number of bootstrap comparisons.
        hill_nboot (int): Number of bootstrap comparisons

    Returns:
        np.ndarray: Hill stats.
        np.ndarray: Sortindex.
        np.ndarray: Upper index.
        np.ndarray: Scan index.
        np.ndarray: Hill data.
        np.ndarray: Hill points.
    """
    indices_ = np.array(query_data["indices_ms1"])
    rt_ = np.array(query_data["rt_list_ms1"])
    mass_data = np.array(query_data["mass_list_ms1"])
    scan_idx = np.searchsorted(indices_, np.arange(len(mass_data)), side="right") - 1
    int_data = np.array(query_data["int_list_ms1"])

    stats = np.zeros((len(hill_ptrs) - 1, 6))  # mz, delta, rt_min, rt_max, sum_max
    hill_stats(
        range(len(hill_ptrs) - 1),
        np.arange(len(hill_ptrs) - 1),
        hill_ptrs,
        hill_data,
        int_data,
        mass_data,
        rt_,
        scan_idx,
        stats,
        hill_nboot_max,
        hill_nboot,
    )

    # sort the stats
    sortindex = np.argsort(stats[:, 4])  # Sorted by rt_min
    stats = stats[sortindex, :]
    idxs_upper = stats[:, 4].searchsorted(stats[:, 5], side="right")
    sortindex_ = np.arange(len(sortindex))[sortindex]

    return stats, sortindex_, idxs_upper, scan_idx, hill_data, hill_ptrs



from typing import List

import numba
import numpy as np
import pandas as pd


def find_spec_idxes_by_rt(
    spectrum_df: pd.DataFrame,
    query_start_rt: float,
    query_stop_rt: float,
    query_left_mz: float,
    query_right_mz: float,
) -> np.ndarray:
    """
    Find MS2 spectrum indices (int32) from the `spectrum_df`
    by given RT window and precursor m/z window.

    Parameters
    ----------
    spectrum_df : pd.DataFrame
        Spectrum dataframe to find spectrum indices.
    query_start_rt : float
        Left RT of the query RT window.
    query_stop_rt : float
        Right RT of the query RT window.
    query_left_mz : float
        Left m/z of the query m/z window.
    query_right_mz : float
        Right m/z of the query m/z window.

    Returns
    -------
    ndarray[int32]
        Result spectrum indices. `int32` is used here as there will be
        no more than 2 billions of spectra in a raw file.
    """
    if "multinotch" in spectrum_df.columns:
        # if multinotch, there are multiple isolation windows of MS2 spectra.
        return find_multinotch_spec_idxes(
            spec_rts=spectrum_df.rt.values,
            spec_multinotch_wins=spectrum_df.multinotch.values,
            spec_ms_levels=spectrum_df.ms_levels.values,
            query_start_rt=query_start_rt,
            query_stop_rt=query_stop_rt,
            query_left_mz=query_left_mz,
            query_right_mz=query_right_mz,
        )
    else:
        # normal isolation windows (one window to one MS2 spectrum)
        return find_spec_idxes(
            spec_rts=spectrum_df.rt.values,
            spec_isolation_lower_mzs=spectrum_df.isolation_lower_mz.values,
            spec_isolation_upper_mzs=spectrum_df.isolation_upper_mz.values,
            query_start_rt=query_start_rt,
            query_stop_rt=query_stop_rt,
            query_left_mz=query_left_mz,
            query_right_mz=query_right_mz,
        )


def find_multinotch_spec_idxes(
    spec_rts: np.ndarray,
    spec_multinotch_wins: List[List],
    spec_ms_levels: np.ndarray,
    query_start_rt: float,
    query_stop_rt: float,
    query_left_mz: float,
    query_right_mz: float,
) -> np.ndarray:
    """
    Find MS2 spectrum indices (int32) from the "multinotch" `spectrum_df`
    by given RT window and precursor m/z window.
    "multinotch" means there are multiple isolation windows of MS2 spectra.

    Parameters
    ----------
    spec_rts : np.ndarray
        RT values of the spectra.
    spec_multinotch_wins : List[List]
        List (num of spectra) of list (multiple isolation windows).
    spec_ms_levels : np.ndarray
        MS levels of the spectra.
    query_start_rt : float
        Left RT of the query RT window.
    query_stop_rt : float
        Right RT of the query RT window.
    query_left_mz : float
        Left m/z of the query m/z window.
    query_right_mz : float
        Right m/z of the query m/z window.

    Returns
    -------
    np.ndarray[int32]
        Result spectrum indices.
    """
    start_idx = np.searchsorted(spec_rts, query_start_rt)
    stop_idx = np.searchsorted(spec_rts, query_stop_rt) + 1
    spec_idxes = []
    for ispec in range(start_idx, stop_idx):
        for win_left, win_right in spec_multinotch_wins[ispec]:
            if spec_ms_levels[ispec] == 1:
                if query_left_mz <= 0:
                    spec_idxes.append(ispec)
            elif max(query_left_mz, win_left) <= min(query_right_mz, win_right):
                spec_idxes.append(ispec)
    return np.array(spec_idxes, dtype=np.int32)


@numba.njit
def find_dia_spec_idxes_same_window(
    spec_rt_values: np.ndarray,
    query_rt_values: np.ndarray,
    max_spec_per_query: int,
) -> np.ndarray:
    """
    For given array of query RT values, find spectrum indices
    from the subset of spectra within the same normal DIA m/z window.
    This function is numba accelerated.

    Parameters
    ----------
    spec_rt_values : np.ndarray
        RT values of given DIA spectra.
    query_rt_values : np.ndarray
        Query RT values.
    max_spec_per_query : int
        Return maximal spectrum indices (scan windows) for the given query.

    Returns
    -------
    ndarray[int32]
        Result spectrum indices with shape (query num, max_spec_per_query).
    """
    rt_idxes = np.searchsorted(spec_rt_values, query_rt_values)

    spec_idxes = np.full((len(rt_idxes), max_spec_per_query), -1, dtype=np.int32)
    n = max_spec_per_query // 2

    for iquery in range(len(rt_idxes)):
        if rt_idxes[iquery] < n:
            spec_idxes[iquery, :] = np.arange(0, max_spec_per_query)
        else:
            spec_idxes[iquery, :] = np.arange(
                rt_idxes[iquery] - n, rt_idxes[iquery] - n + max_spec_per_query
            )
    return spec_idxes


@numba.njit
def find_spec_idxes(
    spec_rts: np.ndarray,
    spec_isolation_lower_mzs: np.ndarray,
    spec_isolation_upper_mzs: np.ndarray,
    query_start_rt: float,
    query_stop_rt: float,
    query_left_mz: float,
    query_right_mz: float,
) -> np.ndarray:
    """
    Find MS2 spectrum indices (int32) from all the spectra
    by given RT window and precursor m/z window.
    This function is numba accelerated.

    Parameters
    ----------
    spec_rts : np.ndarray
        RT values of the spectra.
    spec_isolation_lower_mzs : np.ndarray
        Left m/z values of the isolation windows.
    spec_isolation_upper_mzs : np.ndarray
        Right m/z values of the isolation windows.
    query_start_rt : float
        Left RT of the query RT window.
    query_stop_rt : float
        Right RT of the query RT window.
    query_left_mz : float
        Left m/z of the query m/z window.
    query_right_mz : float
        Right m/z of the query m/z window.

    Returns
    -------
    np.ndarray[int32]
        Result spectrum indices.
    """
    rt_start_idx = np.searchsorted(spec_rts, query_start_rt)
    rt_stop_idx = np.searchsorted(spec_rts, query_stop_rt) + 1

    spec_idxes = []

    for ispec in range(rt_start_idx, rt_stop_idx):
        if max(query_left_mz, spec_isolation_lower_mzs[ispec]) <= min(
            query_right_mz, spec_isolation_upper_mzs[ispec]
        ):
            spec_idxes.append(ispec)
    return np.array(spec_idxes, dtype=np.int32)


@numba.njit
def find_batch_spec_idxes(
    spec_rts: np.ndarray,
    spec_isolation_lower_mzs: np.ndarray,
    spec_isolation_upper_mzs: np.ndarray,
    query_start_rts: np.ndarray,
    query_stop_rts: np.ndarray,
    query_left_mzs: np.ndarray,
    query_right_mzs: np.ndarray,
    max_spec_per_query: int,
) -> np.ndarray:
    """
    Find MS2 spectrum indices (int32) from all the spectra
    by the given batch of RT windows and precursor m/z windows.
    This function is numba accelerated.

    Parameters
    ----------
    spec_rts : np.ndarray
        RT values of the spectra.
    spec_isolation_lower_mzs : np.ndarray
        Left m/z values of the isolation windows.
    spec_isolation_upper_mzs : np.ndarray
        Right m/z values of the isolation windows.
    query_start_rts : np.ndarray
        Left RT values of the query RT windows.
    query_stop_rts : np.ndarray
        Right RT values of the query RT windows.
    query_left_mzs : np.ndarray
        Left m/z values of the query m/z windows.
    query_right_mzs : np.ndarray
        Right m/z values of the query m/z windows.
    max_spec_per_query : int
        Return maximal spectrum indices (scan windows) for the given query.

    Returns
    -------
    ndarray[int32]
        Result spectrum indices with shape (query num, max_spec_per_query).
    """
    rt_start_idxes = np.searchsorted(spec_rts, query_start_rts)
    rt_stop_idxes = np.searchsorted(spec_rts, query_stop_rts) + 1

    spec_idxes = np.full((len(query_left_mzs), max_spec_per_query), -1, dtype=np.int32)
    for iquery in range(len(rt_start_idxes)):
        idx_list = []
        for ispec in range(rt_start_idxes[iquery], rt_stop_idxes[iquery]):
            if max(query_left_mzs[iquery], spec_isolation_lower_mzs[ispec]) <= min(
                query_right_mzs[iquery], spec_isolation_upper_mzs[ispec]
            ):
                idx_list.append(ispec)
        if len(idx_list) > max_spec_per_query:
            spec_idxes[iquery, :] = idx_list[
                len(idx_list) / 2 - max_spec_per_query // 2 : len(idx_list) / 2
                + max_spec_per_query // 2
                + 1
            ]
        else:
            spec_idxes[iquery, : len(idx_list)] = idx_list
    return spec_idxes



import numpy as np
from alphatims.utils import threadpool
from numba import njit


@threadpool
@njit
def connect_centroids_unidirection(
    x: np.ndarray,
    row_borders: np.ndarray,
    connections: np.ndarray,
    scores: np.ndarray,
    centroids: np.ndarray,
    max_gap: int,
    centroid_tol: float,
):
    """Connect centroids.

    Args:
        x (np.ndarray): Index to datapoint. Note that this using the threadpool, so one passes an ndarray.
        row_borders (np.ndarray): Row borders of the centroids array.
        connections (np.ndarray): Connections matrix to store the connections
        scores (np.ndarray):  Score matrix to store the connections
        centroids (np.ndarray): 1D Array containing the masses of the centroids data.
        max_gap (int): Maximum gap when connecting centroids.
        centroid_tol (float): Centroid tolerance.
    """
    for gap in range(max_gap + 1):
        y = x + gap + 1
        if y >= row_borders.shape[0]:
            return

        start_index_f = 0
        if x > 0:
            start_index_f = row_borders[x - 1]

        centroids_1 = centroids[start_index_f : row_borders[x]]
        start_index_b = row_borders[y - 1]
        centroids_2 = centroids[start_index_b : row_borders[y]]

        i = 0
        j = 0

        while (i < len(centroids_1)) & (j < len(centroids_2)):
            mz1, mz2 = centroids_1[i], centroids_2[j]
            diff = mz1 - mz2
            mz_sum = mz1 + mz2
            delta = 2 * 1e6 * abs(diff) / mz_sum

            if delta < centroid_tol and scores[x, i, gap] > delta:
                scores[x, i, gap] = delta
                connections[x, i, gap] = (connections.shape[1] * y) + j

            if diff > 0:
                j += 1
            else:
                i += 1


def find_centroid_connections(
    rowwise_peaks: np.ndarray,
    row_borders: np.ndarray,
    centroids: np.ndarray,
    max_gap: int,
    centroid_tol: float,
):
    """Wrapper function to call connect_centroids_unidirection

    Args:
        rowwise_peaks (np.ndarray): Length of centroids with respect to the row borders.
        row_borders (np.ndarray): Row borders of the centroids array.
        centroids (np.ndarray): Array containing the centroids data.
        max_gap (int): Maximum gap when connecting centroids.
        centroid_tol (float): Centroid tolerance.
    """

    max_centroids = int(np.max(rowwise_peaks))
    spectra_cnt = len(row_borders) - 1

    connections = np.full((spectra_cnt, max_centroids, max_gap + 1), -1, dtype=np.int32)
    score = np.full((spectra_cnt, max_centroids, max_gap + 1), np.inf)

    connect_centroids_unidirection(
        range(len(row_borders)),
        row_borders,
        connections,
        score,
        centroids,
        max_gap,
        centroid_tol,
    )

    score = score[np.where(score < np.inf)]

    score_median = np.median(score)
    score_std = np.std(score)

    del score, max_centroids, spectra_cnt

    c_shape = connections.shape
    from_r, from_c, from_g = np.where(connections >= 0)
    to_r = connections[from_r, from_c, from_g] // c_shape[1]
    to_c = connections[from_r, from_c, from_g] - to_r * c_shape[1]

    del connections, from_g

    return from_r, from_c, to_r, to_c, score_median, score_std


@threadpool(include_progress_callback=False)
@njit
def convert_connections_to_array(
    x: np.ndarray,
    from_r: np.ndarray,
    from_c: np.ndarray,
    to_r: np.ndarray,
    to_c: np.ndarray,
    row_borders: np.ndarray,
    out_from_idx: np.ndarray,
    out_to_idx: np.ndarray,
):
    """Convert integer indices of a matrix to coordinates.

    Args:
        x (np.ndarray): Input index. Note that we are using the performance function so this is a range.
        from_r (np.ndarray): From array with row coordinates.
        from_c (np.ndarray): From array with column coordinates.
        to_r (np.ndarray): To array with row coordinates.
        to_c (np.ndarray): To array with column coordinates.
        row_borders (np.ndarray): Row borders (for indexing).
        out_from_idx (np.ndarray): Reporting array: 1D index from.
        out_to_idx (np.ndarray): Reporting array: 1D index to.
    """
    row = from_r[x]
    col = from_c[x]
    start_index_f = 0
    if row > 0:
        start_index_f = row_borders[row - 1]
    out_from_idx[x] = start_index_f + col

    row = to_r[x]
    col = to_c[x]
    start_index_f = 0
    if row > 0:
        start_index_f = row_borders[row - 1]
    out_to_idx[x] = start_index_f + col


@threadpool(include_progress_callback=False)
@njit
def eliminate_overarching_vertex(
    x: np.ndarray, from_idx: np.ndarray, to_idx: np.ndarray
):
    """Eliminate overacrhing vertex.

    Args:
        x (np.ndarray): Input index. Note that we are using the performance function so this is a range.
        from_idx (np.ndarray): From index.
        to_idx (np.ndarray): To index.
    """
    if x == 0:
        return

    if from_idx[x - 1] == from_idx[x]:
        to_idx[x] = -1


def connect_centroids(
    rowwise_peaks: np.ndarray,
    row_borders: np.ndarray,
    centroids: np.ndarray,
    max_gap: int,
    centroid_tol: float,
) -> (np.ndarray, np.ndarray, float, float):
    """Function to connect centroids.

    Args:
        rowwise_peaks (np.ndarray): Indexes for centroids.
        row_borders (np.ndarray): Row borders (for indexing).
        centroids (np.ndarray): Centroid data.
        max_gap: Maximum gap.
        centroid_tol: Centroid tol for matching centroids.
    Returns:
        np.ndarray: From index.
        np.ndarray: To index.
        float: Median score.
        float: Std deviation of the score.
    """

    from_r, from_c, to_r, to_c, score_median, score_std = find_centroid_connections(
        rowwise_peaks, row_borders, centroids, max_gap, centroid_tol
    )

    from_idx = np.zeros(len(from_r), np.int32)
    to_idx = np.zeros(len(from_r), np.int32)

    convert_connections_to_array(
        range(len(from_r)), from_r, from_c, to_r, to_c, row_borders, from_idx, to_idx
    )

    eliminate_overarching_vertex(range(len(from_idx)), from_idx, to_idx)

    relavent_idx = np.where(to_idx >= 0)
    from_idx = np.take(from_idx, relavent_idx)[0]
    to_idx = np.take(to_idx, relavent_idx)[0]

    del from_r, from_c, to_r, to_c, relavent_idx
    return from_idx, to_idx, score_median, score_std



import typing
import warnings

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from alphatims.bruker import (
    TimsTOF,
)
from plotly.subplots import make_subplots

from .plot_utils import plot_line_tims_fast

warnings.warn(
    "This module will be deprecated or changed in the future releases",
    category=DeprecationWarning,
)


class XIC_Plot_Tims:
    # hovermode = "x" | "y" | "closest" | False | "x unified" | "y unified"
    hovermode = "closest"
    plot_height = 550
    colorscale_qualitative = "Alphabet"
    colorscale_sequential = "Viridis"
    theme_template = "plotly_white"
    ppm = 20.0
    rt_sec_win = 30.0
    plot_rt_unit: str = "minute"
    im_win = 0.05
    fig: go.Figure = None

    def plot(
        self,
        tims_data: TimsTOF,
        query_df: pd.DataFrame,
        view_dim: typing.Literal["rt", "im"] = "rt",
        title: str = "",
        add_peak_area=True,
    ):
        rt_sec = query_df["rt_sec"].values[0]
        im = 0.0 if "im" not in query_df.columns else query_df["im"].values[0]
        if "precursor_mz" not in query_df.columns:
            precursor_mz = 0.0
        else:
            precursor_mz = query_df.precursor_mz.values[0]
        query_masses = query_df.mz.values
        if "intensity" in query_df.columns:
            query_intensities = query_df.intensity.values
        else:
            query_intensities = None
        ion_names = query_df.ion_name.values

        if "color" not in query_df.columns:
            marker_colors = None
        else:
            marker_colors = query_df.color.values

        return self.plot_query_masses(
            tims_data=tims_data,
            query_masses=query_masses,
            query_ion_names=ion_names,
            query_rt_sec=rt_sec,
            query_im=im,
            precursor_mz=precursor_mz,
            marker_colors=marker_colors,
            view_dim=view_dim,
            query_intensities=query_intensities,
            title=title,
            add_peak_area=add_peak_area,
        )

    def _init_plot(self, view_dim="rt"):
        self.fig = make_subplots(
            cols=1,
            shared_xaxes=True,
            x_title=f"RT ({self.plot_rt_unit})" if view_dim == "rt" else "Mobility",
            y_title="intensity",
            vertical_spacing=0.2,
        )
        self.trace: XIC_Plot_Tims = XIC_Trace_Tims(fig=self.fig, row=1)

    def plot_query_masses(
        self,
        tims_data: TimsTOF,
        query_masses: np.ndarray,
        query_ion_names: typing.List[str],
        query_rt_sec: float,
        query_im: float,
        precursor_mz: float,
        marker_colors: list = None,
        view_dim: typing.Literal["rt", "im"] = "rt",
        query_intensities: np.ndarray = None,
        title="",
        add_peak_area=True,
    ):
        self._init_plot(view_dim=view_dim)
        mass_tols = query_masses * 1e-6 * self.ppm
        if marker_colors is None:
            marker_colors = self._get_color_set(len(query_masses))
        self.trace.add_traces(
            tims_data=tims_data,
            query_masses=query_masses,
            mass_tols=mass_tols,
            ion_names=query_ion_names,
            marker_colors=marker_colors,
            query_rt_sec=query_rt_sec,
            query_im=query_im,
            precursor_left_mz=precursor_mz * (1 - 1e-6 * self.ppm),
            precursor_right_mz=precursor_mz * (1 + 1e-6 * self.ppm),
            view_dim=view_dim,
            rt_sec_win=self.rt_sec_win,
            im_win=self.im_win,
            query_intensities=query_intensities,
            add_peak_area=add_peak_area,
        )
        self.fig.update_layout(
            template=self.theme_template,
            title=dict(text=title, yanchor="bottom"),
            # width=width,
            height=self.plot_height,
            hovermode=self.hovermode,
            showlegend=True,
        )
        return self.fig

    def _get_color_set(self, n_query):
        if n_query <= len(getattr(px.colors.qualitative, self.colorscale_qualitative)):
            color_set = getattr(px.colors.qualitative, self.colorscale_qualitative)
        else:
            color_set = px.colors.sample_colorscale(
                self.colorscale_sequential, samplepoints=n_query
            )
        return color_set


class XIC_Trace_Tims:
    label_format = "{ion_name} {mz:.3f}"
    legend_group = "{ion_name}"  # {ion_name}, {mz} or None
    fig: go.Figure
    row: int = 1
    col: int = 1
    plot_rt_unit: str = "minute"

    def __init__(
        self,
        fig: go.Figure,
        row: int = 1,
        col: int = 1,
        plot_rt_unit: str = "minute",
    ):
        self.fig = fig
        self.row = row
        self.col = col
        self.plot_rt_unit = plot_rt_unit

    def add_traces(
        self,
        tims_data: TimsTOF,
        query_masses: np.ndarray,
        mass_tols: np.ndarray,
        ion_names: typing.List[str],
        marker_colors: typing.List,
        query_rt_sec: float,
        query_im: float,
        precursor_left_mz: float = -1.0,
        precursor_right_mz: float = -1.0,
        view_dim: typing.Literal["rt", "im"] = "rt",
        rt_sec_win=30.0,
        im_win=0.05,
        query_intensities: np.ndarray = None,
        add_peak_area=True,
    ) -> go.Figure:
        """Add traces for the query_masses.

        Args:
            tims_data (TimsTOF): AlphaTims TimsTOF object.
            query_masses (np.ndarray): Query masses.
            ion_names (typing.List[str]): Ion names for query_masses.
            marker_colors (typing.List): Colors for each query mass.
            query_rt_sec (float): Query RT in seconds.
            query_im (float): Query ion mobility.
            precursor_mz (float, optional): Precursor mz, 0 means it is MS1. Defaults to 0.0.
            query_intensities (np.ndarray, optional): Query intensities. Defaults to None.

        Returns:
            go.Figure: self.fig.
        """
        (rt_slice, im_slice, prec_mz_slice, view_indices) = get_plotting_slices(
            tims_data=tims_data,
            rt_sec=query_rt_sec,
            rt_sec_win=rt_sec_win,
            im=query_im,
            im_win=im_win,
            precursor_left_mz=precursor_left_mz,
            precursor_right_mz=precursor_right_mz,
            view_dim=view_dim,
        )

        if query_intensities is None:
            query_intensities = np.zeros_like(query_masses)
        else:
            query_intensities /= query_intensities.max()

        for ion_name, query_mass, query_inten, marker_color, mass_tol in zip(
            ion_names, query_masses, query_intensities, marker_colors, mass_tols
        ):
            self._add_one_trace(
                tims_data=tims_data,
                query_mass=query_mass,
                mass_tol=mass_tol,
                rt_slice=rt_slice,
                im_slice=im_slice,
                prec_mz_slice=prec_mz_slice,
                view_indices=view_indices,
                view_dim=view_dim,
                label=self.label_format.format(ion_name=ion_name, mz=query_mass),
                legend_group=self.legend_group.format(ion_name=ion_name),
                marker_color=marker_color,
                add_peak_area=add_peak_area,
            )
            if query_inten > 0:
                self._add_one_trace(
                    tims_data=tims_data,
                    query_mass=query_mass,
                    mass_tol=mass_tol,
                    rt_slice=rt_slice,
                    im_slice=im_slice,
                    prec_mz_slice=prec_mz_slice,
                    view_indices=view_indices,
                    view_dim=view_dim,
                    label=self.label_format.format(ion_name=ion_name, mz=query_mass),
                    legend_group=self.legend_group.format(ion_name=ion_name),
                    marker_color=marker_color,
                    intensity_scale=-query_inten,
                    add_peak_area=add_peak_area,
                )

    def _add_one_trace(
        self,
        tims_data: TimsTOF,
        query_mass: float,
        mass_tol: float,
        rt_slice: slice,
        im_slice: slice,
        prec_mz_slice: slice,
        view_indices: np.ndarray,
        view_dim: str,
        label: str,
        legend_group: str,
        marker_color: str,
        intensity_scale: float = 1.0,
        add_peak_area=True,
    ):
        frag_indices = tims_data[
            rt_slice,
            im_slice,
            prec_mz_slice,
            slice(
                query_mass - mass_tol,
                query_mass + mass_tol,
            ),
            "raw",
        ]
        if len(frag_indices) == 0:
            return
        self.fig.add_trace(
            plot_line_tims_fast(
                tims_data,
                frag_indices,
                view_indices,
                name=label,
                legend_group=legend_group,
                marker_color=marker_color,
                view_dim=view_dim,
                intensity_scale=intensity_scale,
                add_peak_area=add_peak_area,
            ),
            row=self.row,
            col=self.col,
        )


def get_plotting_slices(
    tims_data: TimsTOF,
    rt_sec: float,
    rt_sec_win: float = 30.0,
    im: float = 0.0,
    im_win: float = 0.05,
    precursor_left_mz: float = -1.0,
    precursor_right_mz: float = -1.0,
    view_dim: str = "rt",
):
    """
    Get plotting slices for target queries in TimsTOF data.
    Args:
        tims_data (TimsTOF): AlphaTims TimsTOF object.
        rt_sec (float): Query RT in seconds.
        rt_sec_win (float, optional): Query RT window in seconds. Defaults to 30.0.
        im (float, optional): Query ion mobility, 0 means no mobility dimension. Defaults to 0..
        im_win (float, optional): Ion mobility window. Defaults to 0.05.
        precursor_mz (float, optional): Precursor mz, 0 means it is MS1. Defaults to 0.0.
        ppm (float, optional): PPM tolerance for `precursor_mz`. Defaults to 20.0.
        view_dim (str, optional): View dimension, "rt" or "im". Defaults to "rt"
    """
    rt_slice = slice(rt_sec - rt_sec_win / 2, rt_sec + rt_sec_win / 2)
    im_slice = slice(im - im_win / 2, im + im_win / 2)

    if precursor_left_mz <= 0:
        prec_mz_slice = 0
        raw_indices = tims_data[rt_slice, im_slice, 0, :, "raw"]
    else:
        prec_mz_slice = slice(
            precursor_left_mz,
            precursor_right_mz,
        )
        raw_indices = tims_data[rt_slice, im_slice, prec_mz_slice, :, "raw"]

    if view_dim == "rt":
        view_indices = np.sort(
            np.array(
                list(
                    set(
                        tims_data.convert_from_indices(
                            raw_indices, return_frame_indices=True
                        )["frame_indices"]
                    )
                ),
                dtype=np.int64,
            )
        )
    else:
        view_indices = np.sort(
            np.array(
                list(
                    set(
                        tims_data.convert_from_indices(
                            raw_indices, return_scan_indices=True
                        )["scan_indices"]
                    )
                ),
                dtype=np.int64,
            )
        )

    return rt_slice, im_slice, prec_mz_slice, view_indices



from typing import Callable, Union

import numpy as np
import pandas as pd
from alphatims.utils import threadpool
from numba import njit
from numba.typed import Dict, List

# TODO: Move hardcoded constants
from alpharaw.feature.chem import (
    DELTA_M,
    DELTA_S,
    M_PROTON,
    mass_to_dist,
)


def find_connected_components(edges, min_size=2):
    num_nodes = np.max(edges) + 1
    parent = np.arange(num_nodes)

    def find(x):
        if parent[x] != x:
            parent[x] = find(parent[x])
        return parent[x]

    def union(x, y):
        rootX = find(x)
        rootY = find(y)
        if rootX != rootY:
            parent[rootY] = rootX

    for edge in edges:
        union(edge[0], edge[1])

    for i in range(num_nodes):
        find(i)

    components = {}
    for node in range(num_nodes):
        root = find(node)
        if root in components:
            components[root].append(node)
        else:
            components[root] = [node]

    pre_isotope_patterns = sorted(
        [
            sorted(component)
            for component in components.values()
            if len(component) >= min_size
        ],
        key=len,
        reverse=True,
    )

    return pre_isotope_patterns


@njit
def check_isotope_pattern(
    mass1: float,
    mass2: float,
    delta_mass1: float,
    delta_mass2: float,
    charge: int,
    iso_mass_range: int = 5,
) -> bool:
    """Check if two masses could belong to the same isotope pattern.

    Args:
        mass1 (float): Mass of the first pattern.
        mass2 (float): Mass of the second pattern.
        delta_mass1 (float): Delta mass of the first pattern.
        delta_mass2 (float): Delta mass of the second pattern.
        charge (int): Charge.
        iso_mass_range (int, optional): Mass range. Defaults to 5.

    Returns:
        bool: Flag to see if pattern belongs to the same pattern.
    """
    delta_mass1 = delta_mass1 * iso_mass_range
    delta_mass2 = delta_mass2 * iso_mass_range

    delta_mass = np.abs(mass1 - mass2)

    left_side = np.abs(delta_mass - DELTA_M / charge)
    right_side = np.sqrt((DELTA_S / charge) ** 2 + delta_mass1**2 + delta_mass2**2)

    return left_side <= right_side


@njit
def correlate(
    scans_: np.ndarray, scans_2: np.ndarray, int_: np.ndarray, int_2: np.ndarray
) -> float:
    """Correlate two scans.

    Args:
        scans_ (np.ndarray): Masses of the first scan.
        scans_2 (np.ndarray): Masses of the second scan.
        int_ (np.ndarray): Intensity of the first scan.
        int_2 (np.ndarray): Intensity of the second scan.

    Returns:
        float: Correlation.
    """

    min_one, max_one = scans_[0], scans_[-1]
    min_two, max_two = scans_2[0], scans_2[-1]

    if (
        min_one + 3 > max_two or min_two + 3 > max_one
    ):  # at least an overlap of 3 elements
        corr = 0
    else:
        min_s = min(min_one, min_two)
        max_s = max(max_one, max_two)

        int_one_scaled = np.zeros(int(max_s - min_s + 1))
        int_two_scaled = np.zeros(int(max_s - min_s + 1))

        int_one_scaled[scans_ - min_s] = int_
        int_two_scaled[scans_2 - min_s] = int_2

        corr = np.sum(int_one_scaled * int_two_scaled) / np.sqrt(
            np.sum(int_one_scaled**2) * np.sum(int_two_scaled**2)
        )

    return corr


# %% ../nbs/04_feature_finding.ipynb 29
@njit
def extract_edge(
    stats: np.ndarray,
    idxs_upper: np.ndarray,
    runner: int,
    max_index: int,
    maximum_offset: float,
    iso_charge_min: int = 1,
    iso_charge_max: int = 6,
    iso_mass_range: int = 5,
) -> list:
    """Extract edges.

    Args:
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        idxs_upper (np.ndarray): Upper index for comparing.
        runner (int): Index.
        max_index (int): Unused.
        maximum_offset (float): Maximum offset when comparing edges.
        iso_charge_min (int, optional): Minimum isotope charge. Defaults to 1.
        iso_charge_max (int, optional): Maximum isotope charge. Defaults to 6.
        iso_mass_range (float, optional): Mass search range. Defaults to 5.

    Returns:
        list: List of edges.
    """
    edges = []

    mass1 = stats[runner, 0]
    delta_mass1 = stats[runner, 1]

    for j in range(runner + 1, idxs_upper[runner]):
        mass2 = stats[j, 0]
        if np.abs(mass2 - mass1) <= maximum_offset:
            delta_mass2 = stats[j, 1]
            for charge in range(iso_charge_min, iso_charge_max + 1):
                if check_isotope_pattern(
                    mass1, mass2, delta_mass1, delta_mass2, charge, iso_mass_range
                ):
                    edges.append((runner, j))
                    break

    return edges


@threadpool(include_progress_callback=False)
@njit
def edge_correlation(
    idx: np.ndarray,
    to_keep: np.ndarray,
    sortindex_: np.ndarray,
    pre_edges: np.ndarray,
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
    int_data: np.ndarray,
    scan_idx: np.ndarray,
    cc_cutoff: float,
):
    """Correlates two edges and flag them it they should be kept.

    Args:
        idx (np.ndarray): Input index. Note that we are using the performance function so this is a range.
        to_keep (np.ndarray): Array with indices which edges should be kept.
        sortindex_ (np.ndarray): Sortindex to access the hills from stats.
        pre_edges (np.ndarray): Array with pre edges.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        scan_idx (np.ndarray): Array containing the scan index for a centroid.
        cc_cutoff (float): Cutoff value for what is considered correlating.
    """
    edge = pre_edges[idx, :]

    y = sortindex_[edge[0]]
    start = hill_ptrs[y]
    end = hill_ptrs[y + 1]
    idx_ = hill_data[start:end]
    int_ = int_data[idx_]
    scans_ = scan_idx[idx_]

    con = sortindex_[edge[1]]
    start = hill_ptrs[con]
    end = hill_ptrs[con + 1]
    idx_2 = hill_data[start:end]
    int_2 = int_data[idx_2]
    scans_2 = scan_idx[idx_2]

    if correlate(scans_, scans_2, int_, int_2) > cc_cutoff:
        to_keep[idx] = 1


def get_pre_isotope_patterns(
    stats: np.ndarray,
    idxs_upper: np.ndarray,
    sortindex_: np.ndarray,
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
    int_data: np.ndarray,
    scan_idx: np.ndarray,
    maximum_offset: float,
    iso_charge_min: int = 1,
    iso_charge_max: int = 6,
    iso_mass_range: float = 5,
    cc_cutoff: float = 0.6,
) -> list:
    """Function to extract pre isotope patterns.

    Args:
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        idxs_upper (np.ndarray): Upper index for comparison.
        sortindex_ (np.ndarray): Sortindex to access the hills from stats.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        scan_idx (np.ndarray): Array containing the scan index for a centroid.
        maximum_offset (float): Maximum offset when matching.
        iso_charge_min (int, optional): Minimum isotope charge. Defaults to 1.
        iso_charge_max (int, optional): Maximum isotope charge. Defaults to 6.
        iso_mass_range (float, optional): Mass search range. Defaults to 5.
        cc_cutoff (float, optional): Correlation cutoff. Defaults to 0.6.

    Returns:
        list: List of pre isotope patterns.
    """
    pre_edges = []

    # Step 1
    for runner in range(len(stats)):
        pre_edges.extend(
            extract_edge(
                stats,
                idxs_upper,
                runner,
                idxs_upper[runner],
                maximum_offset,
                iso_charge_min,
                iso_charge_max,
                iso_mass_range,
            )
        )

    to_keep = np.zeros(len(pre_edges), dtype="int")
    pre_edges = np.array(pre_edges)
    edge_correlation(
        range(len(to_keep)),
        to_keep,
        sortindex_,
        pre_edges,
        hill_ptrs,
        hill_data,
        int_data,
        scan_idx,
        cc_cutoff,
    )
    edges = pre_edges[to_keep.nonzero()]

    pre_isotope_patterns = find_connected_components(edges)

    return pre_isotope_patterns


@njit
def check_isotope_pattern_directed(
    mass1: float,
    mass2: float,
    delta_mass1: float,
    delta_mass2: float,
    charge: int,
    index: int,
    iso_mass_range: float,
) -> bool:
    """Check if two masses could belong to the same isotope pattern.

    Args:
        mass1 (float): Mass of the first pattern.
        mass2 (float): Mass of the second pattern.
        delta_mass1 (float): Delta mass of the first pattern.
        delta_mass2 (float): Delta mass of the second pattern.
        charge (int): Charge.
        index (int): Index (unused).
        iso_mass_range (float): Isotope mass ranges.
    Returns:
        bool: Flag if two isotope patterns belong together.
    """
    delta_mass1 = delta_mass1 * iso_mass_range
    delta_mass2 = delta_mass2 * iso_mass_range

    left_side = np.abs(mass1 - mass2 - index * DELTA_M / charge)
    right_side = np.sqrt((DELTA_S / charge) ** 2 + delta_mass1**2 + delta_mass2**2)

    return left_side <= right_side


@njit
def grow(
    trail: List,
    seed: int,
    direction: int,
    relative_pos: int,
    index: int,
    stats: np.ndarray,
    pattern: np.ndarray,
    charge: int,
    iso_mass_range: float,
    sortindex_: np.ndarray,
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
    int_data: np.ndarray,
    scan_idx: np.ndarray,
    cc_cutoff: float,
) -> List:
    """Grows isotope pattern based on a seed and direction.

    Args:
        trail (List): List of hills belonging to a pattern.
        seed (int): Seed position.
        direction (int): Direction in which to grow the trail
        relative_pos (int): Relative position.
        index (int): Index.
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        pattern (np.ndarray): Isotope pattern.
        charge (int): Charge.
        iso_mass_range (float): Mass range for checking isotope patterns.
        sortindex_ (np.ndarray): Sortindex to access the hills from stats.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        scan_idx (np.ndarray): Array containing the scan index for a centroid.
        cc_cutoff (float): Cutoff value for what is considered correlating.

    Returns:
        List: List of hills belonging to a pattern.
    """
    x = pattern[seed]  # This is the seed
    mass1 = stats[x, 0]
    delta_mass1 = stats[x, 1]

    k = sortindex_[x]
    start = hill_ptrs[k]
    end = hill_ptrs[k + 1]
    idx_ = hill_data[start:end]
    int_ = int_data[idx_]
    scans_ = scan_idx[idx_]

    growing = True

    while growing:
        if direction == 1:
            if seed + relative_pos == len(pattern):
                growing = False
                break
        else:
            if seed + relative_pos < 0:
                growing = False
                break

        y = pattern[seed + relative_pos]  # This is a reference peak

        ll = sortindex_[y]

        mass2 = stats[y, 0]
        delta_mass2 = stats[y, 1]

        start = hill_ptrs[ll]
        end = hill_ptrs[ll + 1]
        idx_ = hill_data[start:end]
        int_2 = int_data[idx_]
        scans_2 = scan_idx[idx_]

        if correlate(
            scans_, scans_2, int_, int_2
        ) > cc_cutoff and check_isotope_pattern_directed(
            mass1,
            mass2,
            delta_mass1,
            delta_mass2,
            charge,
            -direction * index,
            iso_mass_range,
        ):
            if direction == 1:
                trail.append(y)
            else:
                trail.insert(0, y)
            index += 1  # Greedy matching: Only one edge for a specific distance, will not affect the following matches

        delta_mass = np.abs(mass1 - mass2)

        if (
            delta_mass > (DELTA_M + DELTA_S) * index
        ):  # the pattern is sorted so there is a maximum to look back
            break

        relative_pos += direction

    return trail


@njit
def grow_trail(
    seed: int,
    pattern: np.ndarray,
    stats: np.ndarray,
    charge: int,
    iso_mass_range: float,
    sortindex_: np.ndarray,
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
    int_data: np.ndarray,
    scan_idx: np.ndarray,
    cc_cutoff: float,
) -> List:
    """Wrapper to grow an isotope pattern to the left and right side.

    Args:
        seed (int): Seed position.
        pattern (np.ndarray): Isotope pattern.
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        charge (int): Charge.
        iso_mass_range (float): Mass range for checking isotope patterns.
        sortindex_ (np.ndarray): Sortindex to access the hills from stats.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        scan_idx (np.ndarray): Array containing the scan index for a centroid.
        cc_cutoff (float): Cutoff value for what is considered correlating.

    Returns:
        List: Isotope pattern.
    """
    x = pattern[seed]
    trail = List()
    trail.append(x)
    trail = grow(
        trail,
        seed,
        -1,
        -1,
        1,
        stats,
        pattern,
        charge,
        iso_mass_range,
        sortindex_,
        hill_ptrs,
        hill_data,
        int_data,
        scan_idx,
        cc_cutoff,
    )
    trail = grow(
        trail,
        seed,
        1,
        1,
        1,
        stats,
        pattern,
        charge,
        iso_mass_range,
        sortindex_,
        hill_ptrs,
        hill_data,
        int_data,
        scan_idx,
        cc_cutoff,
    )

    return trail


@njit
def get_trails(
    seed: int,
    pattern: np.ndarray,
    stats: np.ndarray,
    charge_range: List,
    iso_mass_range: float,
    sortindex_: np.ndarray,
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
    int_data: np.ndarray,
    scan_idx: np.ndarray,
    cc_cutoff: float,
) -> List:
    """Wrapper to extract trails for a given charge range.

    Args:
        seed (int): Seed index.
        pattern (np.ndarray): Pre isotope pattern.
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        charge_range (List): Charge range.
        iso_mass_range (float): Mass range for checking isotope patterns.
        sortindex_ (np.ndarray): Sortindex to access the hills from stats.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        scan_idx (np.ndarray): Array containing the scan index for a centroid.
        cc_cutoff (float): Cutoff value for what is considered correlating.

    Returns:
        List: Trail of consistent hills.
    """
    trails = []
    for charge in charge_range:
        trail = grow_trail(
            seed,
            pattern,
            stats,
            charge,
            iso_mass_range,
            sortindex_,
            hill_ptrs,
            hill_data,
            int_data,
            scan_idx,
            cc_cutoff,
        )

        trails.append(trail)

    return trails


# %% ../nbs/04_feature_finding.ipynb 33
def plot_pattern(
    pattern: np.ndarray,
    sorted_hills: np.ndarray,
    centroids: np.ndarray,
):
    """Helper function to plot a pattern.

    Args:
        pattern (np.ndarray): Pre isotope pattern.
        sorted_hills (np.ndarray): Hills, sorted.
        centroids (np.ndarray): 1D Array containing the masses of the centroids.
        hill_data (np.ndarray): Array containing the indices to hills.
    """
    import matplotlib.pyplot as plt

    f, (ax1, ax2) = plt.subplots(2, 1, sharex=True, figsize=(10, 10))
    centroid_dtype = [("mz", float), ("int", float), ("scan_no", int), ("rt", float)]

    for entry in pattern:
        hill = sorted_hills[entry]
        hill_data = np.array(
            [centroids[_[0]][_[1]] for _ in hill], dtype=centroid_dtype
        )

        ax1.plot(hill_data["rt"], hill_data["int"])
        ax2.scatter(hill_data["rt"], hill_data["mz"], s=hill_data["int"] / 5e5)

    ax1.set_title("Pattern")
    ax1.set_xlabel("RT (min)")
    ax1.set_ylabel("Intensity")

    ax2.set_xlabel("RT (min)")
    ax2.set_ylabel("m/z")

    plt.show()


@njit
def get_minpos(y: np.ndarray, iso_split_level: float) -> List:
    """Function to get a list of minima in a trace.
    A minimum is returned if the ratio of lower of the surrounding maxima to the minimum is larger than the splitting factor.

    Args:
        y (np.ndarray): Input array.
        iso_split_level (float): Isotope split level.

    Returns:
        List: List with min positions.
    """
    minima = get_local_minima(y)
    minima_list = List()

    for minpos in minima:
        minval = y[minpos]

        left_max = (y[:minpos]).max()
        right_max = (y[minpos:]).max()

        minimum_max = min(left_max, right_max)

        if minimum_max / minval >= iso_split_level:
            minima_list.append(minpos)

    return minima_list


@njit
def get_local_minima(y: np.ndarray) -> List:
    """Function to return all local minima of a array

    Args:
        y (np.ndarray): Input array.

    Returns:
        List: List with indices to minima.
    """
    minima = List()
    for i in range(1, len(y) - 1):
        if is_local_minima(y, i):
            minima.append(i)
    return minima


@njit
def is_local_minima(y: np.ndarray, i: int) -> bool:
    """Check if position is a local minima.

    Args:
        y (np.ndarray): Input array.
        i (int): Position to check.

    Returns:
        bool: Flag if position is minima or not.
    """
    return (y[i - 1] > y[i]) & (y[i + 1] > y[i])


@njit
def truncate(
    array: np.ndarray,
    intensity_profile: np.ndarray,
    seedpos: int,
    iso_split_level: float,
) -> np.ndarray:
    """Function to truncate an intensity profile around its seedposition.

    Args:
        array (np.ndarray):  Input array.
        intensity_profile (np.ndarray): Intensities for the input array.
        seedpos (int): Seedposition.
        iso_split_level (float): Split level.

    Returns:
        np.ndarray: Truncated array.
    """
    minima = int_list_to_array(get_minpos(intensity_profile, iso_split_level))

    if len(minima) > 0:
        left_minima = minima[minima < seedpos]
        right_minima = minima[minima > seedpos]

        # If the minimum is smaller than the seed
        minpos = left_minima[-1] if len(left_minima) > 0 else 0

        maxpos = right_minima[0] if len(right_minima) > 0 else len(array)

        array = array[minpos : maxpos + 1]

    return array


@njit
def check_averagine(
    stats: np.ndarray,
    pattern: np.ndarray,
    charge: int,
    averagine_aa: Dict,
    isotopes: Dict,
) -> float:
    """Function to compare a pattern to an averagine model.

    Args:
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        pattern (np.ndarray): Isotope pattern.
        charge (int): Charge.
        averagine_aa (Dict): Dict containing averagine masses.
        isotopes (Dict): Dict containing isotopes.

    Returns:
        float: Averagine correlation.
    """
    masses, intensity = pattern_to_mz(stats, pattern, charge)

    spec_one = np.floor(masses).astype(np.int64)
    int_one = intensity

    spec_two, int_two = mass_to_dist(
        np.min(masses), averagine_aa, isotopes
    )  # maybe change to no rounded version

    spec_two = np.floor(spec_two).astype(np.int64)

    return cosine_averagine(int_one, int_two, spec_one, spec_two)


@njit
def pattern_to_mz(
    stats: np.ndarray, pattern: np.ndarray, charge: int
) -> (np.ndarray, np.ndarray):
    """Function to calculate masses and intensities from pattern for a given charge.

    Args:
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        pattern (np.ndarray): Isotope pattern.
        charge (int): Charge of the pattern.

    Returns:
        np.ndarray: masses
        np.ndarray: intensity
    """

    mzs = np.zeros(len(pattern))
    ints = np.zeros(len(pattern))

    for i in range(len(pattern)):
        entry = pattern[i]
        mzs[i] = mz_to_mass(stats[entry, 0], charge)
        ints[i] = stats[entry, 2]

    sortindex = np.argsort(mzs)

    masses = mzs[sortindex]
    intensity = ints[sortindex]

    return masses, intensity


@njit
def cosine_averagine(
    int_one: np.ndarray, int_two: np.ndarray, spec_one: np.ndarray, spec_two: np.ndarray
) -> float:
    """Calculate the cosine correlation of two hills.

    Args:
        int_one (np.ndarray): Intensity of the first hill.
        int_two (np.ndarray): Intensity of the second hill.
        spec_one (np.ndarray): Scan numbers of the first hill.
        spec_two (np.ndarray): Scan numbers of the second hill.

    Returns:
        float: Cosine
    """

    min_one, max_one = spec_one[0], spec_one[-1]
    min_two, max_two = spec_two[0], spec_two[-1]

    min_s = np.min(np.array([min_one, min_two]))
    max_s = np.max(np.array([max_one, max_two]))

    int_one_scaled = np.zeros(int(max_s - min_s + 1))
    int_two_scaled = np.zeros(int(max_s - min_s + 1))

    int_one_scaled[spec_one - min_s] = int_one
    int_two_scaled[spec_two - min_s] = int_two

    corr = np.sum(int_one_scaled * int_two_scaled) / np.sqrt(
        np.sum(int_one_scaled**2) * np.sum(int_two_scaled**2)
    )

    return corr


@njit
def int_list_to_array(numba_list: List) -> np.ndarray:
    """Numba compatbilte function to convert a numba list with integers to a numpy array

    Args:
        numba_list (List): Input numba-typed List.

    Returns:
        np.ndarray: Output numpy array.
    """
    array = np.zeros(len(numba_list), dtype=np.int64)

    for i in range(len(array)):
        array[i] = numba_list[i]

    return array


@njit
def mz_to_mass(mz: float, charge: int) -> float:
    """Function to calculate the mass from a mz value.

    Args:
        mz (float): M/z
        charge (int): Charge.

    Raises:
        NotImplementedError: When a negative charge is used.

    Returns:
        float: mass
    """
    if charge < 0:
        raise NotImplementedError("Negative Charges not implemented.")

    mass = mz * charge - charge * M_PROTON

    return mass


@njit
def isolate_isotope_pattern(
    pre_pattern: np.ndarray,
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
    int_data: np.ndarray,
    scan_idx: np.ndarray,
    stats: np.ndarray,
    sortindex_: np.ndarray,
    iso_mass_range: float,
    charge_range: List,
    averagine_aa: Dict,
    isotopes: Dict,
    iso_n_seeds: int,
    cc_cutoff: float,
    iso_split_level: float,
) -> (np.ndarray, int):
    """Isolate isotope patterns.

    Args:
        pre_pattern (np.ndarray): Pre isotope pattern.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        scan_idx (np.ndarray): Array containing the scan index for a centroid.
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        sortindex_ (np.ndarray): Sortindex to access the hills from stats.
        iso_mass_range (float): Mass range for checking isotope patterns.
        charge_range (List): Charge range.
        averagine_aa (Dict): Dict containing averagine masses.
        isotopes (Dict): Dict containing isotopes.
        iso_n_seeds (int): Number of seeds.
        cc_cutoff (float): Cutoff value for what is considered correlating.
        iso_split_level (float): Split level when isotopes are split.

    Returns:
        np.ndarray: Array with the best pattern.
        int: Charge of the best pattern.
    """
    longest_trace = 0
    champion_trace = None
    champion_charge = 0
    champion_intensity = 0

    # Sort patterns by mass

    sortindex = np.argsort(stats[pre_pattern][:, 0])  # intensity
    sorted_pattern = pre_pattern[sortindex]
    massindex = np.argsort(stats[sorted_pattern][:, 2])[::-1][:iso_n_seeds]

    # Use all the elements in the pre_pattern as seed

    for seed in massindex:  # Loop through all seeds
        seed_global = sorted_pattern[seed]

        trails = get_trails(
            seed,
            sorted_pattern,
            stats,
            charge_range,
            iso_mass_range,
            sortindex_,
            hill_ptrs,
            hill_data,
            int_data,
            scan_idx,
            cc_cutoff,
        )

        for index, trail in enumerate(trails):
            if (
                len(trail) >= longest_trace
            ):  # Needs to be longer than the current champion
                arr = int_list_to_array(trail)
                intensity_profile = stats[arr][:, 2]
                seedpos = np.nonzero(arr == seed_global)[0][0]

                # truncate around the seed...
                arr = truncate(arr, intensity_profile, seedpos, iso_split_level)
                intensity_profile = stats[arr][:, 2]

                # Remove lower masses:
                # Take the index of the maximum and remove all masses on the left side
                if charge_range[index] * stats[seed_global, 0] < 1000:
                    maxpos = np.argmax(intensity_profile)
                    arr = arr[maxpos:]
                    intensity_profile = stats[arr][:, 2]

                if (len(arr) > longest_trace) | (
                    (len(arr) == longest_trace)
                    & (intensity_profile.sum() > champion_intensity)
                ):
                    # Averagine check
                    cc = check_averagine(
                        stats, arr, charge_range[index], averagine_aa, isotopes
                    )
                    if cc > 0.6:
                        # Update the champion
                        champion_trace = arr
                        champion_charge = charge_range[index]
                        longest_trace = len(arr)
                        champion_intensity = intensity_profile.sum()

    return champion_trace, champion_charge


def get_isotope_patterns(
    pre_isotope_patterns: list,
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
    int_data: np.ndarray,
    scan_idx: np.ndarray,
    stats: np.ndarray,
    sortindex_: np.ndarray,
    averagine_aa: Dict,
    isotopes: Dict,
    iso_charge_min: int = 1,
    iso_charge_max: int = 6,
    iso_mass_range: float = 5,
    iso_n_seeds: int = 100,
    cc_cutoff: float = 0.6,
    iso_split_level: float = 1.3,
    callback: Union[Callable, None] = None,
) -> (np.ndarray, np.ndarray, np.ndarray):
    """Wrapper function to iterate over pre_isotope_patterns.

    Args:
        pre_isotope_patterns (list): List of pre-isotope patterns.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        scan_idx (np.ndarray): Array containing the scan index for a centroid.
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        sortindex_ (np.ndarray): Sortindex to access the hills from stats.
        averagine_aa (Dict): Dict containing averagine masses.
        isotopes (Dict): Dict containing isotopes.
        iso_charge_min (int, optional): Minimum isotope charge. Defaults to 1.
        iso_charge_max (int, optional): Maximum isotope charge. Defaults to 6.
        iso_mass_range (float, optional): Mass search range. Defaults to 5.
        iso_n_seeds (int, optional): Number of isotope seeds. Defaults to 100.
        cc_cutoff (float, optional): Cuttoff for correlation.. Defaults to 0.6.
        iso_split_level (float, optional): Isotope split level.. Defaults to 1.3.
        callback (Union[Callable, None], optional): Callback function for progress. Defaults to None.
    Returns:
        list: List of isotope patterns.
        np.ndarray: Iso idx.
        np.ndarray: Array containing isotope charges.
    """

    isotope_patterns = []
    isotope_charges = []

    charge_range = List()

    for i in range(iso_charge_min, iso_charge_max + 1):
        charge_range.append(i)

    isotope_patterns = []
    isotope_charges = []

    for idx, pre_pattern in enumerate(pre_isotope_patterns):
        extract = True
        while extract:
            isotope_pattern, isotope_charge = isolate_isotope_pattern(
                np.array(pre_pattern),
                hill_ptrs,
                hill_data,
                int_data,
                scan_idx,
                stats,
                sortindex_,
                iso_mass_range,
                charge_range,
                averagine_aa,
                isotopes,
                iso_n_seeds,
                cc_cutoff,
                iso_split_level,
            )
            length = 0 if isotope_pattern is None else len(isotope_pattern)

            if length > 1:
                isotope_charges.append(isotope_charge)
                isotope_patterns.append(isotope_pattern)

                pre_pattern = [_ for _ in pre_pattern if _ not in isotope_pattern]

                if len(pre_pattern) <= 1:
                    extract = False
            else:
                extract = False

        if callback:
            callback((idx + 1) / len(pre_isotope_patterns))

    iso_patterns = np.zeros(sum([len(_) for _ in isotope_patterns]), dtype=np.int64)

    iso_idx = np.zeros(len(isotope_patterns) + 1, dtype="int")

    start = 0
    for idx, _ in enumerate(isotope_patterns):
        iso_patterns[start : start + len(_)] = _
        start += len(_)
        iso_idx[idx + 1] = start

    return iso_patterns, iso_idx, np.array(isotope_charges)


@threadpool
@njit
def report_(
    idx: np.ndarray,
    isotope_charges: list,
    isotope_patterns: list,
    iso_idx: np.ndarray,
    stats: np.ndarray,
    sortindex_: np.ndarray,
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
    int_data: np.ndarray,
    rt_: np.ndarray,
    rt_idx: np.ndarray,
    results: np.ndarray,
    lookup_idx: np.ndarray,
):
    """Function to extract summary statstics from a list of isotope patterns and charges.

    MS1 feature intensity estimation. For each isotope envelope we interpolate the signal over the retention time
    range. All isotope enevelopes are summed up together to estimate the peak sahpe

    Lastly, we report three estimates for the intensity:

    - ms1_int_sum_apex: The intensity at the peak of the summed signal.
    - ms1_int_sum_area: The area of the summed signal
    - ms1_int_max_apex: The intensity at the peak of the most intense isotope envelope
    - ms1_int_max_area: The area of the the most intense isotope envelope

    Args:
        idx (np.ndarray): Input index. Note that we are using the performance function so this is a range.
        isotope_patterns (list): List containing isotope patterns (indices to hills).
        isotope_charges (list): List with charges assigned to the isotope patterns.
        iso_idx (np.ndarray): Index to isotope pattern.
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        sortindex_ (np.ndarray): Sortindex to access the hills from stats.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.
        int_data (np.ndarray): Array containing the intensity to each centroid.
        rt_ (np.ndarray): Array with retention time information for each scan.
        rt_idx (np.ndarray): Lookup array to match centroid idx to rt.
        results (np.ndarray): Recordarray with isotope pattern summary statistics.
        lookup_idx (np.ndarray): Lookup array for each centroid.
    """
    pattern = isotope_patterns[iso_idx[idx] : iso_idx[idx + 1]]
    isotope_data = stats[pattern]

    mz = np.min(isotope_data[:, 0])
    mz_std = np.mean(isotope_data[:, 1])
    charge = isotope_charges[idx]
    mass = mz_to_mass(mz, charge)
    int_max_idx = np.argmax(isotope_data[:, 2])
    mz_most_abundant = isotope_data[:, 0][int_max_idx]

    rt_min_ = min(isotope_data[:, 4])
    rt_max_ = max(isotope_data[:, 5])

    rt_range = np.linspace(
        rt_min_, rt_max_, 100
    )  # TODO this is a fixed value - is there an optimum?

    trace_sum = np.zeros_like(rt_range)

    most_intense_pattern = -np.inf

    for i, k in enumerate(pattern):
        x = sortindex_[k]

        start = hill_ptrs[x]
        end = hill_ptrs[x + 1]
        idx_ = hill_data[start:end]
        int_ = int_data[idx_]
        rts = rt_[rt_idx[idx_]]

        lookup_idx[idx_, 0] = idx
        lookup_idx[idx_, 1] = i

        interpolation = np.interp(rt_range, rts, int_)

        # Filter

        interpolation[: (rt_range < rts[0]).sum()] = 0

        right_cut = (rt_range > rts[-1]).sum()
        if right_cut > 0:
            interpolation[-right_cut:] = 0

        trace_sum += interpolation

        if int_.sum() > most_intense_pattern:
            most_intense_pattern = int_.sum()
            ms1_int_max_apex = int_.max()
            ms1_int_max_area = np.trapz(int_, rts)

    rt_apex_idx = trace_sum.argmax()
    rt_apex = rt_range[rt_apex_idx]

    trace = trace_sum
    half_max = trace.max() / 2

    if rt_apex_idx == 0:
        left_apex = 0
    else:
        left_apex = np.abs(trace[:rt_apex_idx] - half_max).argmin()
    right_apex = np.abs(trace[rt_apex_idx:] - half_max).argmin() + rt_apex_idx

    ms1_int_sum_apex = trace_sum[rt_apex_idx]

    fwhm = rt_range[right_apex] - rt_range[left_apex]

    n_isotopes = len(pattern)

    rt_cutoff = 0.95  # 5%
    if rt_apex_idx == 0:
        rt_min_idx = 0
    else:
        rt_min_idx = np.abs(
            trace[:rt_apex_idx] - trace.max() * (1 - rt_cutoff)
        ).argmin()
    rt_max_idx = (
        np.abs(trace[rt_apex_idx:] - trace.max() * (1 - rt_cutoff)).argmin()
        + rt_apex_idx
    )

    # plt.xlabel('rt')
    # plt.ylabel('int')
    # plt.show()
    # plt.plot(rt_range, trace_sum)

    # plt.plot([rt_range[left_apex], rt_range[right_apex]], [(trace[left_apex] + trace[right_apex])/2]*2, 'k:')

    # plt.plot(rt_range[rt_apex_idx], trace[rt_apex_idx], 'k*')
    # plt.plot(rt_range[rt_min_idx], trace[rt_min_idx], 'k*')
    # plt.plot(rt_range[rt_max_idx], trace[rt_max_idx], 'k*')

    # plt.show()

    rt_start = rt_range[rt_min_idx]
    rt_end = rt_range[rt_max_idx]

    ms1_int_sum_area = np.trapz(
        trace_sum[rt_min_idx:rt_max_idx], rt_range[rt_min_idx:rt_max_idx]
    )

    results[idx, :] = np.array(
        [
            mz,
            mz_std,
            mz_most_abundant,
            charge,
            rt_start,
            rt_apex,
            rt_end,
            fwhm,
            n_isotopes,
            mass,
            ms1_int_sum_apex,
            ms1_int_sum_area,
            ms1_int_max_apex,
            ms1_int_max_area,
        ]
    )


def feature_finder_report(
    query_data: dict,
    isotope_patterns: list,
    isotope_charges: list,
    iso_idx: np.ndarray,
    stats: np.ndarray,
    sortindex_: np.ndarray,
    hill_ptrs: np.ndarray,
    hill_data: np.ndarray,
) -> pd.DataFrame:
    """Creates a report dataframe with summary statistics of the found isotope patterns.

    Args:
        query_data (dict): Data structure containing the query data.
        isotope_patterns (list): List containing isotope patterns (indices to hills).
        isotope_charges (list): List with charges assigned to the isotope patterns.
        iso_idx (np.ndarray): Index to the isotope pattern.
        stats (np.ndarray): Stats array that contains summary statistics of hills.
        sortindex_ (np.ndarray): Sortindex to access the hills from stats.
        hill_ptrs (np.ndarray): Array containing the bounds to the hill_data.
        hill_data (np.ndarray): Array containing the indices to hills.

    Returns:
        pd.DataFrame: DataFrame with isotope pattern summary statistics.
    """
    rt_ = np.array(query_data["rt_list_ms1"])
    indices_ = np.array(query_data["indices_ms1"])
    mass_data = np.array(query_data["mass_list_ms1"])
    rt_idx = np.searchsorted(indices_, np.arange(len(mass_data)), side="right") - 1

    lookup_idx = np.zeros((len(mass_data), 2), dtype=np.int32) - 1

    int_data = np.array(query_data["int_list_ms1"])

    results = np.zeros((len(isotope_charges), 14))

    report_(
        range(len(isotope_charges)),
        isotope_charges,
        isotope_patterns,
        iso_idx,
        stats,
        sortindex_,
        hill_ptrs,
        hill_data,
        int_data,
        rt_,
        rt_idx,
        results,
        lookup_idx,
    )

    df = pd.DataFrame(
        results,
        columns=[
            "mz",
            "mz_std",
            "mz_most_abundant",
            "charge",
            "rt_start",
            "rt_apex",
            "rt_end",
            "fwhm",
            "n_isotopes",
            "mass",
            "ms1_int_sum_apex",
            "ms1_int_sum_area",
            "ms1_int_max_apex",
            "ms1_int_max_area",
        ],
    )

    df.sort_values(["rt_start", "mz"])

    return df, lookup_idx


def get_stats(isotope_patterns, iso_idx, stats):
    columns = ["mz_average", "delta_m", "int_sum", "int_area", "rt_min", "rt_max"]

    stats_idx = np.zeros(iso_idx[-1], dtype=np.int64)
    stats_map = np.zeros(iso_idx[-1], dtype=np.int64)

    start_ = 0
    end_ = 0

    for idx in range(len(iso_idx) - 1):
        k = isotope_patterns[iso_idx[idx] : iso_idx[idx + 1]]
        end_ += len(k)
        stats_idx[start_:end_] = k
        stats_map[start_:end_] = idx
        start_ = end_

    k = pd.DataFrame(stats[stats_idx], columns=columns)

    k["feature_id"] = stats_map

    return k



import numpy as np

from alpharaw.ms_data_base import (
    PEAK_INTENSITY_DTYPE,
    PEAK_MZ_DTYPE,
    MSData_Base,
    index_ragged_list,
    ms_reader_provider,
)


def read_until(file, until):
    lines = []
    while True:
        line = file.readline().strip()
        if line == "" or line.startswith(until):
            break
        else:
            lines.append(line)
    return lines


def find_line(lines, start):
    for line in lines:
        if line.startswith(start):
            return line
    return None


def parse_scan_from_TITLE(mgf_title):
    # raw_name.scan.scan.charge[.xx.xx]
    return int(mgf_title.split(".")[2])


class MGFReader(MSData_Base):
    """MGF Reader (MS2)"""

    def _import(self, _path: str):
        scan_mz_dict = {}
        scan_charge_dict = {}
        masses_list = []
        intens_list = []
        spec_idx_list = []
        scan_list = []
        rt_list = []
        precursor_mz_list = []
        charge_list = []
        self._has_chimeras = False

        f = open(_path) if isinstance(_path, str) else _path  # noqa: SIM115
        while True:
            line = f.readline()
            if not line:
                break
            if line.startswith("BEGIN IONS"):
                lines = read_until(f, "END IONS")
                masses = []
                intens = []
                scan = None
                RT = 0
                precursor_mz = 0
                charge = 0
                for line in lines:
                    if line[0].isdigit():
                        mass, inten = (float(i) for i in line.strip().split())
                        masses.append(mass)
                        intens.append(inten)
                    elif line.startswith("SCAN="):
                        scan = int(line.split("=")[1])
                    elif line.startswith("RTINSECOND"):
                        RT = float(line.split("=")[1]) / 60
                    elif line.startswith("PEPMASS="):
                        precursor_mz = float(line.split("=")[1])
                    elif line.startswith("CHARGE="):
                        charge = int(line.split("=")[1].strip()[:-1])
                if not scan:
                    title = find_line(lines, "TITLE=")
                    scan = parse_scan_from_TITLE(title)
                if scan in scan_mz_dict:
                    scan_mz_dict[scan].append(precursor_mz)
                    scan_charge_dict[scan].append(charge)
                    self._has_chimeras = True
                    continue
                scan_mz_dict[scan] = [precursor_mz]
                scan_charge_dict[scan] = [charge]
                scan_list.append(scan)
                spec_idx_list.append(scan - 1)
                rt_list.append(RT)
                precursor_mz_list.append(precursor_mz)
                charge_list.append(charge)
                masses_list.append(np.array(masses, dtype=PEAK_MZ_DTYPE))
                intens_list.append(np.array(intens, dtype=PEAK_INTENSITY_DTYPE))
        if isinstance(_path, str):
            f.close()

        if self._has_chimeras:
            precursor_mz_list = [scan_mz_dict[scan] for scan in scan_list]
            charge_list = [scan_charge_dict[scan] for scan in scan_list]

        return {
            "peak_indices": index_ragged_list(masses_list),
            "peak_mz": np.concatenate(masses_list),
            "peak_intensity": np.concatenate(intens_list),
            "rt": np.array(rt_list),
            "precursor_mz": precursor_mz_list,
            "spec_idx": np.array(spec_idx_list, dtype=np.int64),
            "scan": np.array(scan_list, dtype=np.int64),
            "precursor_charge": charge_list,
        }

    def _set_dataframes(self, raw_data: dict):
        spec_idxes = raw_data["spec_idx"]
        spec_num = spec_idxes.max() + 1
        self.create_spectrum_df(spec_num)
        start_idxes = np.full(spec_num, -1, dtype=np.int64)
        start_idxes[spec_idxes] = raw_data["peak_indices"][:-1]
        end_idxes = np.full(spec_num, -1, dtype=np.int64)
        end_idxes[spec_idxes] = raw_data["peak_indices"][1:]
        rt_values = np.zeros(spec_num)
        rt_values[spec_idxes] = raw_data["rt"]
        if self._has_chimeras:
            precursor_mzs = [[]] * spec_num
            charges = [[]] * spec_num
            mz_vals = raw_data["precursor_mz"]
            ch_vals = raw_data["precursor_charge"]
            for i, idx in enumerate(spec_idxes):
                precursor_mzs[idx] = mz_vals[i]
                charges[idx] = ch_vals[i]
        else:
            precursor_mzs = np.zeros(spec_num)
            precursor_mzs[spec_idxes] = raw_data["precursor_mz"]
            charges = np.zeros(spec_num, np.int8)
            charges[spec_idxes] = raw_data["precursor_charge"]

        self.spectrum_df["charge"] = charges
        self.spectrum_df["precursor_mz"] = precursor_mzs

        self.set_peak_df_by_indexed_array(
            raw_data["peak_mz"], raw_data["peak_intensity"], start_idxes, end_idxes
        )
        self.add_column_in_spec_df_by_spec_idxes(
            "rt", raw_data["rt"], spec_idxes, na_value=0
        )
        self.add_column_in_spec_df_by_spec_idxes(
            "ms_level", 2, spec_idxes, dtype=np.int8, na_value=1
        )


def register_readers():
    """
    Register :class:`MGFReader` for file format "mgf" in
    :obj:`alpharaw.ms_data_base.ms_reader_provider`.
    """
    ms_reader_provider.register_reader("mgf", MGFReader)



from __future__ import annotations

import numpy as np
from numba import float32, float64, int32, njit, types
from numba.experimental import jitclass
from numba.typed import Dict

# Some legacy chemistry functions
# Maybe replacd with alphabase?


averagine_aa = Dict.empty(key_type=types.unicode_type, value_type=types.float64)

averagine_aa["C"] = 4.9384
averagine_aa["H"] = 7.7583
averagine_aa["N"] = 1.3577
averagine_aa["O"] = 1.4773
averagine_aa["S"] = 0.0417

averagine_avg = 111.1254

DELTA_M = 1.00286864
DELTA_S = 0.0109135

M_PROTON = 1.00727646687

maximum_offset = DELTA_M + DELTA_S


spec = [
    ("m0", float32),
    ("dm", int32),
    ("intensities", float32[:]),
]


@jitclass(spec)
class Isotope:
    """
    Jit-compatible class to store isotopes

    Attributes:
        m0 (int): Mass of pattern
        dm0 (int): dm of pattern (number of isotopes)
        int0 (np.float32[:]): Intensities of pattern
    """

    def __init__(self, m0: int, dm: int, intensities: np.ndarray):
        self.m0 = m0
        self.dm = dm
        self.intensities = intensities


isotopes = Dict.empty(
    key_type=types.unicode_type, value_type=Isotope.class_type.instance_type
)

isotopes["C"] = Isotope(12, 3, np.array([0.9893, 0.0107, 0.0], dtype=np.float32))
isotopes["H"] = Isotope(
    1.007940, 3, np.array([0.999885, 0.000115, 0.0], dtype=np.float32)
)
isotopes["O"] = Isotope(
    15.9949146221, 3, np.array([0.99757, 0.00038, 0.00205], dtype=np.float32)
)
isotopes["N"] = Isotope(
    14.0030740052, 2, np.array([0.99636, 0.00364], dtype=np.float32)
)
isotopes["S"] = Isotope(
    31.97207069, 4, np.array([0.9499, 0.0075, 0.0425, 0.0001], dtype=np.float32)
)

isotopes["I"] = Isotope(126.904473, 1, np.array([1], dtype=np.float32))
isotopes["K"] = Isotope(
    38.9637069, 3, np.array([0.932581, 0.000117, 0.067302], dtype=np.float32)
)


def test_isotope():
    assert isotopes["C"].m0 == 12
    assert isotopes["C"].dm == 3
    assert np.allclose(isotopes["C"].intensities[0], 0.9893)
    assert np.allclose(isotopes["C"].intensities[1], 0.0107)
    assert np.allclose(isotopes["C"].intensities[2], 0)


# test_isotope()


spec = [
    ("m0", float32),
    ("dm", int32),
    ("intensities", float64[:]),
]


# This class and the methods fast_add, numba_bin are tested in the wrapper function dict_to_dist
@jitclass(spec)
class IsotopeDistribution:
    """Class to represent isotope distributions.

    Attributes:
        m0 (int): the mono-isotopic mass.
        dm (int): number of isotopes.
        intensities (np.ndarray): isotope intensities.

    Methods:
        add: add another isotope distribution.
        copy: create a copy of the current isotope distribution.
        mult: multiply the current isotope distribution.

    """

    def __init__(self):
        self.m0 = 0
        self.dm = 1
        self.intensities = np.array([1.0])

    def add(self, x: IsotopeDistribution):
        """Add another isotope distribution.

        Args:
            x (IsotopeDistribution): IsotopeDistribution.
        """
        self.m0, self.dm, self.intensities = fast_add(
            self.m0, self.dm, self.intensities, x.m0, x.dm, x.intensities
        )

    def copy(self) -> IsotopeDistribution:
        """Copies the current isotope distribution.

        Returns:
            IsotopeDistribution: A copy of the current isotope distribution.
        """
        i = IsotopeDistribution()
        i.m0 = self.m0
        i.dm = self.dm
        i.intensities = self.intensities

        return i

    def mult(self, n: int) -> IsotopeDistribution:
        """Multiplies the current isotope distribution.

        Args:
            n (int): Multiplication factor.

        Returns:
            IsotopeDistribution: A multiplied version of the current isotope distribution.
        """
        binary = numba_bin(n)

        if n == 1:
            return self.copy()
        else:
            i = IsotopeDistribution()

            multiples = self.copy()

            for count in binary[::-1]:
                if count == 1:
                    i.add(multiples)
                multiples.add(multiples)

            return i


@njit
def fast_add(
    m0: int,
    dm0: int,
    int0: np.ndarray,
    m1: int,
    dm1: int,
    int1: np.ndarray,
    prune_level: float = 0.000001,
) -> (int, int, np.ndarray):
    """Helper function to quickly add isotope patterns.

    Args:
        m0 (float): the mono-isotopic mass of the first isotope distribution.
        dm0 (int): the number of isotope intensities in the first isotope distribution.
        int0 (np.ndarray:float): the intensity pattern of the first isotope distribution.
        m1 (float): the mono-isotopic mass of the second isotope distribution.
        dm1 (int): the number of isotope intensities in the second isotope distribution.
        int1 (np.ndarray:float): the intensity pattern of the second isotope distribution.
        prune_level (float): Precision threshold. Defaults to 0.000001.


    Returns:
        int: Mass of new pattern.
        int: Number of isotopes in new pattern.
        np.ndarray: Intensity of new pattern.

    """
    m0 += m1

    ni = np.zeros(dm0 + dm1 - 1)
    for i in range(dm0):
        for j in range(dm1):
            ni[i + j] += int0[i] * int1[j]

    dm0 += dm1 - 1

    int0 = ni / np.max(ni)

    while ni[dm0 - 1] < prune_level:
        dm0 -= 1

    return m0, dm0, int0


@njit
def numba_bin(decimal: int) -> list:
    """Numba compatible function to convert a decimal number to a binary (list).

    Args:
        decimal (int): Decimal number.

    Returns:
        list: Number in binary.
    """

    binary = []

    while decimal != 0:
        bit = int(decimal % 2)
        binary.insert(0, bit)
        decimal = int(decimal / 2)

    return binary


@njit
def dict_to_dist(counted_AA: Dict, isotopes: Dict) -> IsotopeDistribution:
    """Function to convert a dictionary with counts of atoms to an isotope distribution.

    Args:
        counted_AA (Dict): Numba-typed dict with counts of atoms.
        isotopes (Dict): Numba-typed lookup dict with isotopes.

    Returns:
        IsotopeDistribution: The calculated isotope distribution for the chemical compound.
    """

    dist = IsotopeDistribution()
    for AA in counted_AA:
        x = IsotopeDistribution()
        x.add(isotopes[AA])
        x = x.mult(counted_AA[AA])

        dist.add(x)

    return dist


def test_dict_to_dist():
    abundances = (
        np.array(
            [
                58.83,
                0.18,
                100,
                0.29,
                81.29,
                0.22,
                42.05,
                0.11,
                15.54,
                0.04,
                4.36,
                0.01,
                0.97,
                0.17,
                0.03,
            ]
        )
        / 100
    )

    counted_AA = Dict.empty(key_type=types.unicode_type, value_type=types.int64)
    counted_AA["K"] = 23
    counted_AA["I"] = 22
    dist = dict_to_dist(counted_AA, isotopes)

    rmse = np.mean(np.sqrt((dist.intensities[: len(abundances)] - abundances) ** 2))

    assert rmse < 0.05


# test_dict_to_dist()


@njit
def get_average_formula(
    molecule_mass: float, averagine_aa: Dict, isotopes: Dict, sulphur: bool = True
) -> Dict:
    """Function to calculate the averagine formula for a molecule mass.

    Args:
        molecule_mass (float): Input molecule mass for which the averagine model should be calculated.
        averagine_aa (Dict): Numba-typed dictionary with averagine masses. See alphapept.constants.averagine_aa.
        isotopes (Dict): Numba-typed lookup dict with isotopes.
        sulphur (bool, optional): Flag to consider sulphur. Defaults to True.

    Raises:
        NotImplementedError: If mode w/o sulphur is selected.

    Returns:
        Dict: Numba-typed dict with averagine composition.
    """

    if sulphur:
        averagine_units = molecule_mass / averagine_avg
    else:
        raise NotImplementedError("Mode w/o sulphur is not implemented yet")

    counted_AA = Dict.empty(key_type=types.unicode_type, value_type=types.int64)

    final_mass = 0

    # Calculate integral mnumbers of atoms
    for AA in averagine_aa:
        counted_AA[AA] = int(np.round(averagine_units * averagine_aa[AA]))
        final_mass += counted_AA[AA] * isotopes[AA].m0

    # Correct with H atoms
    h_correction = int(np.round((molecule_mass - final_mass) / isotopes["H"].m0))
    counted_AA["H"] += h_correction

    return counted_AA


def test_get_average_formula():
    for molecule_mass in [100, 200, 300, 400, 500]:
        average_formula = get_average_formula(
            molecule_mass, averagine_aa, isotopes, sulphur=True
        )
        mass = np.sum([average_formula[AA] * isotopes[AA].m0 for AA in average_formula])
        assert np.abs(mass - molecule_mass) < isotopes["H"].m0


# test_get_average_formula()

ISOTOPE_MASS = DELTA_M


@njit
def mass_to_dist(
    molecule_mass: float, averagine_aa: Dict, isotopes: Dict
) -> (np.ndarray, np.ndarray):
    """Function to calculate an isotope distribution from a molecule mass using the averagine model.

    Args:
        molecule_mass (float, averagine_aa): input molecule mass.
        averagine_aa (Dict): Numba-typed dictionary with averagine masses.
        isotopes (Dict): Numba-typed lookup dict with isotopes.

    Returns:
        np.ndarray: isotope masses.
        np.ndarray: isotope intensity.
    """
    counted_AA = get_average_formula(molecule_mass, averagine_aa, isotopes)

    dist = dict_to_dist(counted_AA, isotopes)

    masses = np.array(
        [dist.m0 + i * ISOTOPE_MASS for i in range(len(dist.intensities))]
    )
    ints = dist.intensities

    return masses, ints


# print(mass_to_dist(300, averagine_aa, isotopes))



import typing

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from alphabase.constants.atom import MASS_ISOTOPE
from plotly.subplots import make_subplots

from alpharaw.match.match_utils import match_batch_spec
from alpharaw.match.spec_finder import (
    find_multinotch_spec_idxes,
    find_spec_idxes,
)

from .plot_utils import plot_line


class XIC_Plot:
    """
    Only apply for 3D-MS, i.e. without mobility
    """

    hovermode = "closest"
    """
    hovermode = "x", "y", "closest", False, "x unified", "y unified"
    """

    plot_height = 550
    colorscale_qualitative = "Alphabet"
    colorscale_sequential = "Viridis"
    theme_template = "plotly_white"
    ms1_ppm = 20.0
    ms2_ppm = 20.0
    rt_sec_win = 60.0
    plot_rt_unit: str = "second"
    fig: go.Figure = None
    isotope_cum_abundance = (0.9,)

    # list of XIC_Trace objects
    traces: list = []

    def plot(
        self,
        spectrum_df: pd.DataFrame,
        peak_df: pd.DataFrame,
        query_df: pd.DataFrame,
        title: str = "",
        add_peak_area=False,
        create_new_fig=True,
        plot_rows=1,
        ith_plot_row=0,
    ) -> go.Figure:
        """
        Main entry of `XIC_Plot`.

        Parameters
        ----------
        spectrum_df : pd.DataFrame
            The spectrum dataframe in alpharaw format
        peak_df : pd.DataFrame
            The peak dataframe in alpharaw format
        query_df : pd.DataFrame
            The query dataframe generated by:
            :func:`alpharaw.viz.df_utils.make_query_df_for_peptide`, or
            :func:`alpharaw.viz.df_utils.make_query_df`
        title : str, optional
            Plot title, by default ""
        add_peak_area : bool, optional
            If add peak area in the hover text, by default False
        create_new_fig : bool, optional
            If create a new figure, otherwise plot in `self.fig`, by default True
        plot_rows : int, optional
            If create_new_fig, how many rows (subplots) to create, by default 1
        ith_plot_row : int, optional
            If not create_new_fig, this parameter refers to the row number
            in `self.fig`, by default 0

        Returns
        -------
        go.Figure
            XIC plot figure
        """
        if "rt_sec" in query_df.columns:
            rt_sec = query_df.rt_sec.values[0]
        else:
            rt_sec = query_df.rt.values[0] * 60

        (precursor_left_mz, precursor_right_mz) = self._get_precursor_mz_range(query_df)

        query_masses = query_df.mz.values
        if "intensity" in query_df.columns:
            query_intensities = query_df.intensity.values
        else:
            query_intensities = None
        ion_names = query_df.ion_name.values

        if "color" not in query_df.columns:
            marker_colors = None
        else:
            marker_colors = query_df.color.values

        if add_peak_area:
            self.get_peak_area(spectrum_df, peak_df, query_df)
            other_plot_infos = [
                f"Area: {_area:.3e}" for _area in query_df.peak_area.values
            ]
        else:
            other_plot_infos = None

        return self.plot_query_masses(
            spectrum_df=spectrum_df,
            peak_df=peak_df,
            query_masses=query_masses,
            query_ion_names=ion_names,
            query_rt_sec=rt_sec,
            precursor_left_mz=precursor_left_mz,
            precursor_right_mz=precursor_right_mz,
            marker_colors=marker_colors,
            other_plot_infos=other_plot_infos,
            query_intensities=query_intensities,
            title=title,
            create_new_fig=create_new_fig,
            plot_rows=plot_rows,
            ith_plot_row=ith_plot_row,
        )

    def _get_precursor_mz_range(
        self,
        query_df,
    ):
        if "precursor_mz" not in query_df.columns:
            precursor_left_mz = -1.0
            precursor_right_mz = -1.0
        else:
            precursor_left_mz = query_df.precursor_mz.values[0] * (
                1 - self.ms1_ppm * 1e-6
            )
            precursor_right_mz = query_df.precursor_mz.values[0] * (
                1 + self.ms1_ppm * 1e-6
            )

            for iso in range(10):
                if f"precursor_i_{iso}" not in query_df.columns:
                    break
            if iso > 0:
                mono_idx = query_df.precursor_mono_idx.values[0]
                charge = query_df.precursor_charge.values[0]
                precursor_mz = query_df.precursor_mz.values[0]
                precursor_left_mz = precursor_mz - mono_idx * MASS_ISOTOPE / charge
                isotope_cumsum = np.cumsum(
                    [query_df[f"precursor_i_{i}"].values[0] for i in range(iso)]
                )
                for i in range(iso - 1, -1, -1):
                    if isotope_cumsum[i] >= self.isotope_cum_abundance:
                        precursor_right_mz = (
                            precursor_mz + (i - mono_idx) * MASS_ISOTOPE / charge
                        )
                        break
        return precursor_left_mz, precursor_right_mz

    def get_peak_area(
        self,
        spectrum_df: pd.DataFrame,
        peak_df: pd.DataFrame,
        query_df: pd.DataFrame,
        query_start_rt_sec: float = None,
        query_stop_rt_sec: float = None,
        precursor_left_mz: float = None,
        precursor_right_mz: float = None,
    ):
        if query_start_rt_sec is None:
            query_start_rt_sec = query_df.rt_sec.values[0] - self.rt_sec_win / 2
            query_stop_rt_sec = query_df.rt_sec.values[0] + self.rt_sec_win / 2
        if precursor_left_mz is None:
            (precursor_left_mz, precursor_right_mz) = self._get_precursor_mz_range(
                query_df
            )

        mass_tols = (
            query_df.mz.values
            * 1e-6
            * (self.ms1_ppm if precursor_left_mz <= 0 else self.ms2_ppm)
        )

        query_df["peak_area"] = get_peak_area(
            spectrum_df=spectrum_df,
            peak_df=peak_df,
            query_masses=query_df.mz.values,
            mass_tols=mass_tols,
            query_start_rt_sec=query_start_rt_sec,
            query_stop_rt_sec=query_stop_rt_sec,
            precursor_left_mz=precursor_left_mz,
            precursor_right_mz=precursor_right_mz,
        )
        return query_df

    def plot_query_masses(
        self,
        spectrum_df: pd.DataFrame,
        peak_df: pd.DataFrame,
        query_masses: np.ndarray,
        query_ion_names: typing.List[str],
        query_rt_sec: float,
        precursor_left_mz: float,
        precursor_right_mz: float,
        marker_colors: list = None,
        other_plot_infos: typing.List[str] = None,
        query_intensities: np.ndarray = None,
        title="",
        create_new_fig=True,
        plot_rows=1,
        ith_plot_row=0,
    ) -> go.Figure:
        """
        The other entry of `XIC_Plot`

        Parameters
        ----------
        spectrum_df : pd.DataFrame
            The spectrum dataframe in alpharaw format
        peak_df : pd.DataFrame
            The peak dataframe in alpharaw format
        query_masses : np.ndarray
            The query m/z values to extract XIC profiles
        query_ion_names : typing.List[str]
            The query ion names
        query_rt_sec : float
            RT in seconds
        precursor_left_mz : float
            Left m/z of precursor window for given queries to extract XIC
        precursor_right_mz : float
            Right m/z of precursor window for given queries to extract XIC
        marker_colors : list, optional
            Colors of each XIC profile, by default None
        other_plot_infos : typing.List[str], optional
            Other plot information for hover text, by default None
        query_intensities : np.ndarray, optional
            Query intensities for mirror plot, by default None
        title : str, optional
            Plot title, by default ""
        create_new_fig : bool, optional
            If create a new figure, otherwise plot in `self.fig`, by default True
        plot_rows : int, optional
            If create_new_fig, how many rows (subplots) to create, by default 1
        ith_plot_row : int, optional
            If not create_new_fig, this parameter refers to the row number
            in `self.fig`, by default 0

        Returns
        -------
        go.Figure
            XIC plot figure
        """
        if create_new_fig:
            self._init_plot(rows=plot_rows)
        mass_tols = (
            query_masses
            * 1e-6
            * (self.ms1_ppm if precursor_left_mz < 0 else self.ms2_ppm)
        )
        if marker_colors is None:
            marker_colors = self._get_color_set(len(query_masses))
        self.traces[ith_plot_row].add_traces(
            spectrum_df=spectrum_df,
            peak_df=peak_df,
            query_masses=query_masses,
            mass_tols=mass_tols,
            ion_names=query_ion_names,
            marker_colors=marker_colors,
            query_start_rt_sec=query_rt_sec - self.rt_sec_win / 2,
            query_stop_rt_sec=query_rt_sec + self.rt_sec_win / 2,
            precursor_left_mz=precursor_left_mz,
            precursor_right_mz=precursor_right_mz,
            query_intensities=query_intensities,
            other_plot_infos=other_plot_infos,
        )
        self.fig.update_layout(
            template=self.theme_template,
            title=dict(text=title, yanchor="bottom"),
            # width=width,
            height=self.plot_height,
            hovermode=self.hovermode,
            showlegend=True,
        )
        return self.fig

    def _init_plot(self, rows=1):
        self.fig = make_subplots(
            rows=rows,
            cols=1,
            shared_xaxes=True,
            x_title=f"RT ({self.plot_rt_unit})",
            y_title="intensity",
            vertical_spacing=0.2 / rows,
        )
        self.traces: typing.List[XIC_Trace] = [
            XIC_Trace(
                fig=self.fig,
                row=i + 1,
                plot_rt_unit=self.plot_rt_unit,
            )
            for i in range(rows)
        ]

    def _get_color_set(self, n_query):
        if n_query <= len(getattr(px.colors.qualitative, self.colorscale_qualitative)):
            color_set = getattr(px.colors.qualitative, self.colorscale_qualitative)
        else:
            color_set = px.colors.sample_colorscale(
                self.colorscale_sequential, samplepoints=n_query
            )
        return color_set


class XIC_Trace:
    label_format = "{ion_name} {mz:.3f}"
    legend_group = "{ion_name}"  # {ion_name}, {mz} or None
    fig: go.Figure
    row: int = 1
    col: int = 1
    plot_rt_unit: str = "second"

    def __init__(
        self,
        fig: go.Figure,
        row: int = 1,
        col: int = 1,
        plot_rt_unit: str = "second",
    ):
        self.fig = fig
        self.row = row
        self.col = col
        self.plot_rt_unit = plot_rt_unit

    def add_traces(
        self,
        spectrum_df: pd.DataFrame,
        peak_df: pd.DataFrame,
        query_masses: np.ndarray,
        mass_tols: np.ndarray,
        ion_names: typing.List[str],
        marker_colors: typing.List,
        query_start_rt_sec: float,
        query_stop_rt_sec: float,
        precursor_left_mz: float = -1.0,
        precursor_right_mz: float = -1.0,
        query_intensities: np.ndarray = None,
        other_plot_infos: typing.List[str] = None,
    ) -> go.Figure:
        spec_idxes = get_spec_idxes_from_df(
            spectrum_df,
            query_start_rt_sec=query_start_rt_sec,
            query_stop_rt_sec=query_stop_rt_sec,
            precursor_left_mz=precursor_left_mz,
            precursor_right_mz=precursor_right_mz,
        )

        rt_values = spectrum_df.rt.values[spec_idxes]
        if self.plot_rt_unit == "second":
            rt_values = rt_values * 60
        matched_mzs, matched_intens = match_batch_spec(
            spec_idxes,
            peak_mzs=peak_df.mz.values,
            peak_intens=peak_df.intensity.values,
            peak_start_idxes=spectrum_df.peak_start_idx.values,
            peak_stop_idxes=spectrum_df.peak_stop_idx.values,
            query_mzs=query_masses,
            query_mz_tols=mass_tols,
        )

        if query_intensities is None:
            query_intensities = np.zeros_like(query_masses)
        else:
            query_intensities /= query_intensities.max()

        for i_query, (ion_name, query_mass, query_inten, marker_color) in enumerate(
            zip(ion_names, query_masses, query_intensities, marker_colors)
        ):
            if other_plot_infos is not None:
                other_info = other_plot_infos[i_query]
            else:
                other_info = ""
            if (matched_intens[:, i_query] == 0).all():
                continue
            self._add_one_trace(
                rt_values,
                matched_intens[:, i_query],
                label=self.label_format.format(ion_name=ion_name, mz=query_mass),
                legend_group=self.legend_group.format(ion_name=ion_name),
                marker_color=marker_color,
                other_info=other_info,
            )
            if query_inten > 0:
                self._add_one_trace(
                    rt_values,
                    -query_inten * matched_intens[:, i_query],
                    label=self.label_format.format(ion_name=ion_name, mz=query_mass),
                    legend_group=self.legend_group.format(ion_name=ion_name),
                    marker_color=marker_color,
                    other_info=other_info,
                )

    def _add_one_trace(
        self,
        rt_values: np.ndarray,
        matched_intens: np.ndarray,
        label: str,
        legend_group: str,
        marker_color: str,
        other_info: str = "",
    ):
        self.fig.add_trace(
            plot_line(
                rt_values,
                matched_intens,
                name=label,
                legend_group=legend_group,
                marker_color=marker_color,
                x_text="RT",
                other_info=other_info,
            ),
            row=self.row,
            col=self.col,
        )


def get_spec_idxes_from_df(
    spectrum_df: pd.DataFrame,
    query_start_rt_sec,
    query_stop_rt_sec,
    precursor_left_mz,
    precursor_right_mz,
):
    if "multinotch" in spectrum_df.columns:
        return find_multinotch_spec_idxes(
            spectrum_df.rt.values,
            spectrum_df.multinotch.values,
            spectrum_df.ms_level.values,
            query_start_rt=query_start_rt_sec / 60,
            query_stop_rt=query_stop_rt_sec / 60,
            query_left_mz=precursor_left_mz,
            query_right_mz=precursor_right_mz,
        )
    else:
        return find_spec_idxes(
            spectrum_df.rt.values,
            spectrum_df.isolation_lower_mz.values,
            spectrum_df.isolation_upper_mz.values,
            query_start_rt=query_start_rt_sec / 60,
            query_stop_rt=query_stop_rt_sec / 60,
            query_left_mz=precursor_left_mz,
            query_right_mz=precursor_right_mz,
        )


def get_peak_area(
    spectrum_df: pd.DataFrame,
    peak_df: pd.DataFrame,
    query_masses: np.ndarray,
    mass_tols: np.ndarray,
    query_start_rt_sec: float,
    query_stop_rt_sec: float,
    precursor_left_mz: float = -1.0,
    precursor_right_mz: float = -1.0,
) -> np.ndarray:
    """
    Get peak area values for each of query in `query_masses`

    Parameters
    ----------
    spectrum_df : pd.DataFrame
        The spectrum dataframe
    peak_df : pd.DataFrame
        The peak dataframe
    query_masses : np.ndarray
        The query m/z values
    mass_tols : np.ndarray
        The query m/z tolerance in Da
    query_start_rt_sec : float
        Start RT in seconds
    query_stop_rt_sec : float
        Stop RT in seconds
    precursor_left_mz : float
        Left m/z of precursor window for given queries to extract XIC
    precursor_right_mz : float
        Right m/z of precursor window for given queries to extract XIC

    Returns
    -------
    np.ndarray
        peak area values for each of query
    """
    spec_idxes = get_spec_idxes_from_df(
        spectrum_df,
        query_start_rt_sec=query_start_rt_sec,
        query_stop_rt_sec=query_stop_rt_sec,
        precursor_left_mz=precursor_left_mz,
        precursor_right_mz=precursor_right_mz,
    )

    rt_values = spectrum_df.rt.values[spec_idxes]

    matched_mzs, matched_intens = match_batch_spec(
        spec_idxes,
        peak_mzs=peak_df.mz.values,
        peak_intens=peak_df.intensity.values,
        peak_start_idxes=spectrum_df.peak_start_idx.values,
        peak_stop_idxes=spectrum_df.peak_stop_idx.values,
        query_mzs=query_masses,
        query_mz_tols=mass_tols,
    )

    return np.trapz(y=matched_intens, x=rt_values * 60, axis=0)



import warnings

import alpharaw.raw_access.pysciexwifffilereader as pysciexwifffilereader

from .ms_data_base import MSData_Base, ms_reader_provider


class SciexWiffData(MSData_Base):
    """
    Load Sciex Wiff data as :class:`alpharaw.ms_data_base.MSData_Base` data structure.
    This reader will be registered as "sciex", "sciex_wiff", and "sciex_raw"
    in :obj:`alpharaw.ms_data_base.ms_reader_provider` by :func:`register_readers()`.
    """

    def __init__(self, centroided: bool = True, save_as_hdf: bool = False, **kwargs):
        """
        Parameters
        ----------
        centroided : bool, optional
            If peaks will be centroided after loading,
            by default True.

        save_as_hdf : bool, optional
            Automatically save hdf after load raw data, by default False.
        """
        super().__init__(centroided, save_as_hdf=save_as_hdf, **kwargs)
        if self.centroided:
            self.centroided = False
            warnings.warn("Centroiding for Sciex data is not well implemented yet")
        self.centroid_ppm = 20.0
        self.ignore_empty_scans = True
        self.keep_k_peaks_per_spec = 2000
        self.sample_id = 0
        self.file_type = "sciex"

    def _import(self, _wiff_file_path: str) -> dict:
        """
        Implementation of :func:`alpharaw.ms_data_base.MSData_Base._import` interface.

        Parameters
        ----------
        _wiff_file_path : str
            Absolute or relative path of the sciex wiff file.

        Returns
        -------
        dict
            Spectrum information dict.
        """
        wiff_reader = pysciexwifffilereader.WillFileReader(_wiff_file_path)
        data_dict = wiff_reader.load_sample(
            self.sample_id,
            centroid=self.centroided,
            centroid_ppm=self.centroid_ppm,
            ignore_empty_scans=self.ignore_empty_scans,
            keep_k_peaks=self.keep_k_peaks_per_spec,
        )
        self.creation_time = (
            wiff_reader.wiffSample.Details.AcquisitionDateTime.ToString("O")
        )
        wiff_reader.close()
        return data_dict


def register_readers():
    """
    Register :class:`SciexWiffData` for file formats (types):
    "sciex", "sciex_wiff", and "sciex_raw" in :obj:`alpharaw.ms_data_base.ms_reader_provider`.
    """
    ms_reader_provider.register_reader("sciex", SciexWiffData)
    ms_reader_provider.register_reader("sciex_wiff", SciexWiffData)
    ms_reader_provider.register_reader("sciex_raw", SciexWiffData)
